#include "pch.h"
#include "IntDLL.h"  
#include "IntApi.h"
#include "SoftwareConfig.h"
#include "DebugOutput.h"



extern PLX_DEVICE_OBJECT hPlxDevice;

//
//		Name:			TCR1000_ApiGetDLLInitStatus()
//
//		Description:	Returns DLL initialization status
//

DWORD __stdcall TCR1000_ApiGetDLLInitStatus(void)
{
	DWORD InitStatus;

	AcquireApiMutex();
	InitStatus = dllInitStatus;
	ReleaseApiMutex();
	return(InitStatus);
}

//
//		Name:			TCR1000_ApiGetTimeCodeStatus()
//
//		Description:	Returns current status of time code, one of DLL_TC_STATUS_xxx
//

DWORD __stdcall TCR1000_ApiGetTimeCodeStatus(void)
{
	U8			tcStatus;
	PLX_STATUS  rc;

	AcquireApiMutex();

	//rc = TCRRead(TCR_READ_TC_STATUS, &tcStatus, 1, BitSize8);
	rc = PlxPci_PciBarSpaceReadDynamic(&hPlxDevice, 2, TCR_READ_TC_STATUS, &tcStatus, 1, BitSize8, FALSE);

	if (rc != ApiSuccess)
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!! TCRGet TC Status: TCRread failed: %d",rc);
		#endif

		dwTcType = TC_TYPE_UNDEFINED;

		ReleaseApiMutex();
		return DLL_TC_STATUS_DRVFAIL;
	} else
	{
		#ifdef _DEBUG_INFO1
		DebugMsg("TCRGet TC Status successful");
		#endif
	}

	dwTcType = GetTCTypeFromTCStatus(tcStatus);

	if (tcStatus == 0xFF)
	{
		#ifdef _DEBUG_INFO
		DebugMsg("TCRGet status: calibrating: %d",(int)tcStatus);
		#endif

		ReleaseApiMutex();
		return DLL_TC_STATUS_CALIB;
	}

	if ((tcStatus & 0x80) != 0)
	{
		#ifdef _DEBUG_INFO
		DebugMsg("TCRGet status: lock lost: %d",(int)tcStatus);
		#endif

		ReleaseApiMutex();
		return DLL_TC_STATUS_LOST;
	}

	ReleaseApiMutex();
	return DLL_TC_STATUS_NORMAL;
}


//
//		Name:			TCR1000_ApiGetDeviceStatus()
//
//		Description:	Returns current status of TCR device, one of HW_STATUS_xxx
//

DWORD __stdcall TCR1000_ApiGetDeviceStatus(void)
{
	// Read the status from the card

	U8		   devStatus;
	PLX_STATUS rc;

	AcquireApiMutex();

	//rc = TCRRead(TCR_READ_TC_STATUS, &devStatus, 1, BitSize8);
	rc = PlxPci_PciBarSpaceReadDynamic(&hPlxDevice, 2, TCR_READ_TC_STATUS, &devStatus, 1, BitSize8, FALSE);

	if (rc == ApiSuccess)
	{
		ReleaseApiMutex();
		return HW_STATUS_NORMAL;
	}
	else
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!TCRGet device status: read failed: %d",(int)rc);
		#endif

		ReleaseApiMutex();
		return HW_STATUS_CARDCOMMFAIL;
	}
}

//
//		Name:			TCR1000_ApiSetFrameRate()
//
//		Description:	Sets sub-second frame rate for time code type being decoded.
//						Used to derive wMilliseconds member of TCRTIME
//
//		Returns:		TRUE if set successfully
//						FALSE if specified rate is invalid for current TCRSync-configured time code type
//
//		Note: for TCR1000 and later, this value is now ignored. The frame rate comes directly
//            from the card. Therefore, this function is deprecated.

BOOL __stdcall TCR1000_ApiSetFrameRate(WORD rate)
{
	AcquireApiMutex();

	switch(dwTcType)
	{
		case TC_TYPE_SMPTE:
			if ((rate == 30) || (rate == 25) || (rate == 24))
			{
				ReleaseApiMutex();
				return(TRUE);
			}

			#ifdef _DEBUG_ERROR
			DebugMsg("!!!TCRsetFrameRate: invalid rate:%d",(int)rate);
			#endif

		    ReleaseApiMutex();
			return FALSE;
			break;
	}		

	#ifdef _DEBUG_ERROR
	DebugMsg("!!!TCRsetFrameRate: invalid dwTcType:%d",(int)dwTcType);
	#endif

	ReleaseApiMutex();
	return(FALSE);
}


//
//		Name:			TCR1000_ApiPollTimeCode()
//
//		Description:	Polls the TCR for one time code frame.  This will be accurate to +/- one count
//						of the time code's frame rate.  Not recommended for high-precision timing.
//
//		Returns:		Time code status, one of DLL_TC_STATUS_xxx from INTAPI.H
//						

DWORD __stdcall TCR1000_ApiPollTimeCode(PTCRTIME tcrt)
{
	DWORD dwReturn;
	TIMEFRAME	timeFrame;

	// Read the data from the card

	LARGE_INTEGER tmp;

	AcquireApiMutex();

	tmp.QuadPart = 0;

	dwReturn = PopulateTimeFrame(&timeFrame, tmp);

	if (dwReturn)
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!TCR1000_ApiPollTimeCode: populate fail: %d", (int)dwReturn);
		#endif

		ReleaseApiMutex();
		return DLL_TC_STATUS_DRVFAIL;
	}
		
	// Data read, now process it into the correct format

	ProcessTimeFrame(&timeFrame, tcrt);

	if (timeFrame.tcStatus == 0xFF)	// still calibrating
	{
		#ifdef _DEBUG_INFO
		DebugMsg("TCR1000_ApiPollTimeCode: calibrating");
		#endif

		ReleaseApiMutex();
		return DLL_TC_STATUS_CALIB;
	}

	if ((timeFrame.tcStatus & 0x80) != 0)
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!TCR1000_ApiPollTimeCode: lost lock: %d", (int)timeFrame.tcStatus);
		#endif

		ReleaseApiMutex();
		return DLL_TC_STATUS_LOST;
	}

	ReleaseApiMutex();
	return(DLL_TC_STATUS_NORMAL);
}


//
//		Name:			TCR1000_ApiEnableCallback()
//
//		Description:	Adds or creates asynchronous notification callback capability.  If the calling
//						thread has already registered a callback function, we add the requested
//						asynchronous notification to the current capacity.  Otherwise, we create a new
//						callback function node.
//
//		Returns:		On success, the new asynchronous notification mask
//						On failure, 0
//						

DWORD __stdcall TCR1000_ApiEnableCallback(DWORD cbToAdd, PTCRASYNC_CALLBACK_ROUTINE cbFunction)
{
	PCALLBACK_NODE	cbNode, cbListTail;
	DWORD			curThreadId;
	DWORD			dwHooksMask;

	AcquireApiMutex();

	// only one thread in this process may manipulate the callback function list
	// at any one time.
	dwHooksMask = 0;

	__try
	{

		// synchronize with other threads

		#ifdef _DEBUG_INFO
		DebugMsg("Enable callback critical section start");
		#endif
		EnterCriticalSection(&g_srDllShared.csCBListData);

		// find (or create) a callback node to manipulate

		curThreadId = GetCurrentThreadId();
		cbNode = cbListHead;

		while (cbNode && cbNode->cbThreadId != curThreadId)
		{
			cbListTail = cbNode;
			cbNode = (PCALLBACK_NODE)cbNode->next;
		}

		if (!cbNode)
		{
			// add a new callback function node for this process

			if (!cbListHead)
			{
				// the first node in our list

				cbListHead = (PCALLBACK_NODE)malloc(sizeof(CALLBACK_NODE));
				cbListHead->next = NULL;
				cbListHead->prev = NULL;
				cbNode = cbListHead;
			}
			else
			{
				// a new node at the end of the list

				cbNode = (PCALLBACK_NODE)malloc(sizeof(CALLBACK_NODE));
				cbNode->prev = cbListTail;
				cbListTail->next = cbNode;
			}

			// initialize the new node

			cbNode->cbHooksMask = 0;
			cbNode->locked = FALSE;

			cbNode->cbThreadId = curThreadId;

			DuplicateHandle(GetCurrentProcess(),
							GetCurrentThread(),
							GetCurrentProcess(),
							&cbNode->cbThreadHandle,
							NULL,
							FALSE,
							DUPLICATE_SAME_ACCESS);
		}

		if (!cbNode->cbHooksMask)
		{
			// this is the first callback request for calling thread, set callback function address

			cbNode->cbFunction = cbFunction;
		}

		// Add the hook to this call back, if it's not already there

		if (!(cbNode->cbHooksMask & cbToAdd))
		{

			cbNode->cbHooksMask |= cbToAdd;
			dwHooksMask = cbNode->cbHooksMask;

			// Increment the CB counters to determine the interrupt status beging

			AddCallbackCounts(cbToAdd);
		}
	}

	__finally
	{
		#ifdef _DEBUG_INFO
		DebugMsg("Enable callback critical section end");
		#endif

		LeaveCriticalSection(&g_srDllShared.csCBListData);

	}

	ReleaseApiMutex();
	return dwHooksMask;
}


//
//						

//DWORD __clrcall TCRTestClr(void)
//{
//	AcquireApiMutex();
//	ReleaseApiMutex();
//	return 1;
//}


//
//		Name:			TCR1000_ApiDisableCallback()
//
//		Description:	Disables one or more hooks to a thread's callback function.  When all hooks are
//						cleared the callback function is no longer called and is removed from the callback
//						list.
//
//		Returns:		On success, new hook mask
//						On failure, ASYNC_MASK_ERROR
//						

DWORD __stdcall TCR1000_ApiDisableCallback(DWORD cbToRemove)
{
	DWORD			curThreadId;
	PCALLBACK_NODE	cbNode;
	DWORD			dwHooksMask;

	AcquireApiMutex();

	// only one thread in this process may manipulate the callback function list
	// at any one time.
	dwHooksMask = 0;

	__try
	{

		// synchronize with other threads

		#ifdef _DEBUG_INFO
		DebugMsg("Disable callback critical section start");
		#endif
		EnterCriticalSection(&g_srDllShared.csCBListData);

		curThreadId = GetCurrentThreadId();
		cbNode = cbListHead;

		// Look for the node that matches the current thread
		
		while (cbNode && (cbNode->cbThreadId != curThreadId))
		{
			cbNode = (PCALLBACK_NODE)cbNode->next;
		}

		// If no node matches the thread, or there is no matching callback, then exit.

		if (!cbNode || !(cbNode->cbHooksMask & cbToRemove))
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!TCR1000_ApiDisableCallback: callback not found");
			#endif
			dwHooksMask = ASYNC_MASK_ERROR;
		}
		else
		{
			cbNode->cbHooksMask &= ~cbToRemove;

			if (!cbNode->cbHooksMask)
			{
				// all hooks have been disabled so remove this callback node from the list

				CloseHandle(cbNode->cbThreadHandle);

				// maintain head and list pointers
				
				if (cbNode == cbListHead)
					cbListHead = (PCALLBACK_NODE)cbNode->next;
				else
					((PCALLBACK_NODE)cbNode->prev)->next = cbNode->next;	// route around current node in list

				free(cbNode); // remove node
				cbNode = NULL;
			}

			if (cbNode) 
				dwHooksMask = cbNode->cbHooksMask;
			else
				dwHooksMask = 0;

			// Decrememnt the callback counters (which disables the interrupt if necessary)
			RemoveCallbackCounts(cbToRemove);
		}
	}

	__finally
	{
		#ifdef _DEBUG_INFO
		DebugMsg("Disable callback critical section end");
		#endif

		LeaveCriticalSection(&g_srDllShared.csCBListData);
	}

	ReleaseApiMutex();
	return dwHooksMask;
}


//
//		Name:			TCR1000_ApiShutdown()
//
//		Description:	Should be called before a _process_ (using an instance of the DLL)
//						terminates.  Allows DLL to gracefully terminate threads and deallocate
//						resources on behalf of the calling process.

void __stdcall TCR1000_ApiShutdown(void)
{
	AcquireApiMutex();
	#ifdef _DEBUG_INFO
	DebugMsg("TCR1000_ApiShutdown: shutdown");
	#endif

	ReleaseApiMutex();
	return;	// now this does nothing
}


//
//		Name:			TCR1000_ApiClearEnhanced()
//
//		Description:	Clears enhanced features (once/frame and once/second interrupts) on the TCR-200/500.
//
//		Note:			This routine violates the multiprocess/multithreaded premise of this DLL.
//						A better mechanism should be implemented eventually.
//
//		Returns:		TRUE on completion
//						

BOOL __stdcall TCR1000_ApiClearEnhanced(void)
{
	AcquireApiMutex();
	TCR1000_ApiDisableCallback(ASYNC_ENDTC);
	TCR1000_ApiDisableCallback(ASYNC_STARTTC);
	TCR1000_ApiDisableCallback(ASYNC_PERSEC);

	ReleaseApiMutex();
	return TRUE;
}


//
//		Name:			TCR1000_ApiRequestRecalibration()
//
//		Description:	Initiates a time code signal level calibration on the TCR-500/500PCI
//
//		Returns:		TRUE upon completion, or FALSE if not available or an error
//

BOOL __stdcall TCR1000_ApiRequestRecalibration(void)
{
	RETURN_CODE rc;
	U8 val = 0;

	AcquireApiMutex();

	//rc = TCRWrite(TCR_WRITE_RECALIBRATE, &val, 1, BitSize8);
	rc = PlxPci_PciBarSpaceWriteDynamic(&hPlxDevice, 2, TCR_WRITE_RECALIBRATE, &val, 1, BitSize8, FALSE);
	if (rc == ApiSuccess)
	{
		#ifdef _DEBUG_INFO
		DebugMsg("TCR1000_ApiRequestRecalibration: recalibration started");
		#endif
		ReleaseApiMutex();
		return TRUE;
	}
	else
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!TCR1000_ApiRequestRecalibration: recalibration request failed: %d", (int)rc);
		#endif
		ReleaseApiMutex();
		return FALSE;
	}
}

//
//		Name:			TCR1000_ApiGetTimeCodeType()
//
//		Description:	Returns time code type currently in use/configured by TCRSync
//
//		Returns:		One of TC_TYPE_xxx defined in INTAPI.H
//
	
DWORD __stdcall TCR1000_ApiGetTimeCodeType(void)
{
	// since this function calls another API function which uses the API mutex,
	// no APi mutex calls are required in this function

	//DebugBreak();
	TCR1000_ApiGetTimeCodeStatus(); // populates timeCodeType

	return(dwTcType);
}


//
//		Name:			TCR1000_ApiGetGpsStatus()
//
//		Description:	Returns Time code lock status if SMPTE, Leitch date encoding, with Masterclock, Inc proprietary control bits
//
//		Returns:		One of GPS_STATUS_xxx defined in INTAPI.H
//
	
DWORD __stdcall TCR1000_ApiGetGpsStatus(void)
{
	DWORD Status;

	AcquireApiMutex();
	Status = devInfo.gpsStatus;
	ReleaseApiMutex();

	return(Status);
}


//
//		Name:			TCR1000_ApiGetDLLVersion()
//
//		Description:	Returns DLL version number
//

void __stdcall TCR1000_ApiGetDLLVersion(PUINT verMaj, PUINT verMin, PUINT verRev)
{
	AcquireApiMutex();
	if (verMaj)
		*verMaj = DLL_VERSION_MAJ;

	if (verMin)
		*verMin = DLL_VERSION_MIN;

	if (verRev)
		*verRev = DLL_VERSION_REV;

	ReleaseApiMutex();
}

//
//		Name:			TCR1000_ApiGetDriverVersion()
//
//		Description:	Returns TCR driver version
//
//		Returns:		TRUE if DLL initialized correctly, with driver version in parameters
//

BOOL __stdcall TCR1000_ApiGetDriverVersion(PUINT verMaj, PUINT verMin, PUINT verRev)
{
	PLX_STATUS rc;
	U8	driverVersionMajor;
	U8	driverVersionMinor;
	U8	driverVersionRevision;

	AcquireApiMutex();

	if (dllInitStatus == DLL_INIT_OK)
	{
		rc = PlxPci_DriverVersionDynamic(&hPlxDevice, &driverVersionMajor, &driverVersionMinor, &driverVersionRevision);

		if (rc == ApiSuccess)
		{
			
			if (verMaj)
				*verMaj = (UINT)driverVersionMajor;		// firmware stores values as chars

			if (verMin)
				*verMin = (UINT)driverVersionMinor;

			if (verRev)
				*verRev = (UINT)driverVersionRevision;

			ReleaseApiMutex();
			return TRUE;
		} else
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!TCR1000_ApiGetDriverVersion: read failed: %d",(int)rc);
			#endif

			if (verMaj) *verMaj = 0;
			if (verMin) *verMin = 0;
			if (verRev) *verRev = 0;

			ReleaseApiMutex();
			return FALSE;
		}

	}
	else
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!TCR1000_ApiGetDriverVersion: dll init failed: %d", (int)dllInitStatus);
		#endif

		if (verMaj) *verMaj = 0;
		if (verMin) *verMin = 0;
		if (verRev) *verRev = 0;

		ReleaseApiMutex();
		return FALSE;
	}
}




//
//		Name:			TCR1000_ApiSetConfiguration()
//
//		Description:	Sets the configuration and saves it in the registry.
//
//		Returns:		0 if successful
//                      else 32-bit bitmap indicating which parameters were invalid
//
UINT32 __stdcall TCR1000_ApiSetConfiguration(TCR1000_CONFIGURATION *pApiSetConfig)
{
	UINT32 ulretval;

	AcquireApiMutex();

	CSoftwareConfig	 *pcSaveConfig = new CSoftwareConfig();

	ulretval = pcSaveConfig->SaveConfigFromAPI(pApiSetConfig);

	if (!ulretval)
	{
		pcSaveConfig->SaveConfigToRegistry();
	}

	delete pcSaveConfig;

	ReleaseApiMutex();
	return ulretval;
}




//
//		Name:			TCR1000_ApiGetConfiguration()
//
//		Description:	Gets the current configuration.
//
//		Returns:		0 if successful
//                      else 32-bit bitmap indicating which parameters were invalid
//
void __stdcall TCR1000_ApiGetConfiguration(TCR1000_CONFIGURATION *pApiSetConfig)
{

	AcquireApiMutex();
	CSoftwareConfig	 *pcGetConfig = new CSoftwareConfig();

	pcGetConfig->GetConfigForAPI(pApiSetConfig);

	delete pcGetConfig;
	ReleaseApiMutex();
}








//
//		Name:			TCR1000_ApiGetFirmwareVersion()
//
//		Description:	Returns TCR firmware version
//
//		Returns:		TRUE if value retrieved successfully, FALSE if not
//
BOOL __stdcall TCR1000_ApiGetFirmwareVersion(PUINT verMaj, PUINT verMin, PUINT verRev)
{
	PLX_STATUS rc;
	U8	fwVer[3];

	AcquireApiMutex();

	if (dllInitStatus == DLL_INIT_OK)
	{
		//rc = TCRReadApi(0x0D, fwVer, 3, BitSize8);
		rc = PlxPci_PciBarSpaceReadDynamic(&hPlxDevice, 2, 0x0D, fwVer, 3, BitSize8, FALSE);
		
		if (rc == ApiSuccess)
		{
			if (verMaj)
				*verMaj = (UINT) fwVer[0] - '0';

			if (verMin)
				*verMin = (UINT) fwVer[1] - '0';

			if (verRev)
				*verRev = (UINT) fwVer[2] - '0';

			ReleaseApiMutex();
			return TRUE;
		} else
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!TCR1000_ApiGetFirmwareVersion: read failed: %d", (int)rc);
			#endif

			if (verMaj) *verMaj = 0;
			if (verMin) *verMin = 0;
			if (verRev) *verRev = 0;

			ReleaseApiMutex();
			return FALSE;
		}
	}
	else
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!TCR1000_ApiGetFirmwareVersion: dll init failed: %d", (int)dllInitStatus);
		#endif

		if (verMaj) *verMaj = 0;
		if (verMin) *verMin = 0;
		if (verRev) *verRev = 0;

		ReleaseApiMutex();
		return FALSE;
	}
}

//
//		Name:			TCR1000_ApiGetFirmwareChecksum()
//
//		Description:	Returns TCR firmware checksum
//
//		Returns:		TRUE if value retrieved successfully, FALSE if not
//

BOOL __stdcall TCR1000_ApiGetFirmwareChecksum(PDWORD pChecksum)
{

	DWORD		checksum = 0xdeadbeef;
	PLX_STATUS rc;

	AcquireApiMutex();
	if (dllInitStatus == DLL_INIT_OK)
	{
		//rc = TCRReadApi(0x10, &checksum, 4, BitSize32);
		rc = PlxPci_PciBarSpaceReadDynamic(&hPlxDevice, 2, 0x10, &checksum, 4, BitSize32, FALSE);

		if (rc == ApiSuccess)
		{
			if (pChecksum)
			{
				*pChecksum = checksum;				

				ReleaseApiMutex();				
				return TRUE;
			} else
			{
				ReleaseApiMutex();
				return FALSE;
			}
		}
		else
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!TCR1000_ApiGetFirmwareChecksum: read failed: %d", (int)rc);
			#endif

			ReleaseApiMutex();
			return FALSE;
		}

	}

	#ifdef _DEBUG_ERROR
	DebugMsg("!!!TCR1000_ApiGetFirmwareChecksum: dll init failed: %d",(int)dllInitStatus);
	#endif

	ReleaseApiMutex();
	return FALSE;
}
