
#include "pch.h"

#include "IntDLL.h"
#include "tzlib.h"
#include "DebugOutput.h"

#define ASYNC_TIMING_INTS	(ASYNC_TIMESET | ASYNC_PERSEC | ASYNC_STARTTC | ASYNC_ENDTC)

//MC_DLL_SHARED g_srDllShared = {0};

// interprocess DLL instance shared data

#pragma data_seg(".shared")

// Sharing is not working as expected!!! When loaded, the DLL gives each
// process it's own copy of the shared memory. ie. sharing is doing nothing
MC_DLL_SHARED g_srDllShared = {0};

#pragma data_seg()

//#pragma comment(linker, "/section:.shared,RWS") // this line is not needed, see the .def file
//#pragma comment(linker, "/section:.data,RW")    // this line is not needed, see the .def file


// private DLL functions

static DWORD ReadTimeFrameFromPlx(TIMEFRAME *timeFrame);
static DWORD InitDLLInstance(void);
static DWORD WINAPI AsyncEventMonitor(LPVOID lpParam);
static DWORD WriteInterruptStatus(U8 intsToSet);
static DWORD GetInterruptData(PUCHAR pIntReason, PTIMEFRAME pTimeFrame);
static DWORD GetPollingTimeframe(PTIMEFRAME pTimeFrame);
static RETURN_CODE EnableTCRInterrupt(DWORD intToEnable);
static RETURN_CODE DisableTCRInterrupt(DWORD intToEnable);

static boolean fnLoadPlxDll(HINSTANCE *hPlxDllInstanceLib);
static boolean fnFreePlxDll(HINSTANCE hinstLib);

// private DLL data members

//DEVICE_LOCATION	device;
PLX_DEVICE_OBJECT hPlxDevice;
PLX_PARAMS        plxIoBuffer;

U8				tcStatus;
U8              tcBuffer[8];		// Time code data takes 8 bytes
U8              tcChecksum;
//HANDLE			tcrInterruptEvent;
DEVICE_INFO		devInfo;
DWORD			dllInitStatus;
DWORD			dwTcType = TC_TYPE_UNDEFINED;

HANDLE			hTempTcr;
PCALLBACK_NODE	cbListHead = NULL;
DWORD			asyncMonitorId;

BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved)
{
	DWORD			curThreadId;
	PCALLBACK_NODE	cbNode, cbPrevNode;
	HINSTANCE      *indirect;
	boolean         bIsSuccessful;

	#ifdef _DEBUG_INFO
	DebugMsg("MC API DLL main: %d", (int)dwReason);
	#endif

	switch (dwReason)
    {
		case DLL_PROCESS_ATTACH:
			// called when a new application (process) first uses the DLL
			
			// this pointer is located in the shared data area.
			// We only do the load one time

			if (!g_srDllShared.hPlxDllInstanceLib)
			{
				#ifdef _DEBUG_OUTPUT_ENABLE
				DebugInit("TCR.DLL", "C:\\TEMP\\TCRDEBUG.TXT", DEBUG_FILE);
				#endif

				indirect = &g_srDllShared.hPlxDllInstanceLib;

				bIsSuccessful = fnLoadPlxDll(indirect);

				if (!bIsSuccessful)
				{
					// make sure the pointer is still a NULL since the load failed
					g_srDllShared.hPlxDllInstanceLib = NULL;

					#ifdef _DEBUG_ERROR
					DebugMsg("!!!Unable to load PLX DLL");
					#endif
				}

				#ifdef _DEBUG_INIT
				if (bIsSuccessful)
				{
					DebugMsg("PLX DLL loaded successfully");
				}
				#endif
			} else
			{
				#ifdef _DEBUG_INIT
					DebugMsg("PLX DLL was already loaded successfully");
				#endif
			}

			InitializeCriticalSection(&g_srDllShared.csCBListData);

			// we can only run the init if the DLL was successfully loaded
			if (g_srDllShared.hPlxDllInstanceLib)
			{
				dllInitStatus = InitDLLInstance();
			} else
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Plx DLL not loaded, could not run init instance");
				#endif
			}

			#ifdef _DEBUG_INFO
			DebugMsg("DLL process attach critical section start");
			#endif
			EnterCriticalSection(&g_srDllShared.csCBListData);

			g_srDllShared.dllInstanceCnt++;

			#ifdef _DEBUG_INFO
			DebugMsg("DLL process attach critical section end");
			#endif
			LeaveCriticalSection(&g_srDllShared.csCBListData);


			#ifdef _DEBUG_INFO
			DebugMsg("DLL_PROCESS_ATTACH - DLL instance count: %d",(int)g_srDllShared.dllInstanceCnt);
			#endif

			break;
        
		case DLL_PROCESS_DETACH:
			// called when an application (process) using the DLL terminates, note that all threads
			// spawned by the DLL have already been terminated (by the OS)

			// check callback list, and remove all the callbacks for
			// all threads that are left


			#ifdef _DEBUG_INFO
			DebugMsg("DLL process detach critical section start");
			#endif

			EnterCriticalSection(&g_srDllShared.csCBListData);
			
			cbNode = cbListHead;

			while (cbNode)
			{
				// delete this callback node since it will no longer be needed
				#ifdef _DEBUG_INFO
				DebugMsg("Deleting a callback node because process terminated w/o disabling its callback.");
				#endif

				// Adjust the expanded interrupt counts, based on removing
				// the callbacks associated with this thread

				RemoveCallbackCounts(cbNode->cbHooksMask);

				cbPrevNode = cbNode;
				
				cbNode = (PCALLBACK_NODE)cbNode->next;

				free(cbPrevNode); // remove node
			}

			cbListHead = NULL;	// All modes deleted

			g_srDllShared.dllInstanceCnt--;

			#ifdef _DEBUG_INFO
			DebugMsg("DLL_PROCESS_DETACH - DLL instance count: %lu", (unsigned long)g_srDllShared.dllInstanceCnt);
			#endif

			if (g_srDllShared.dllInstanceCnt == 0)
			{
				#ifdef _DEBUG_INFO
				DebugMsg("The DLL will be leaving memory.");
				#endif

				// Close the device


				#ifdef _DEBUG_INFO
				DebugMsg("PLX DLL will be leaving memory.");
				#endif
				if (g_srDllShared.hPlxDllInstanceLib)
				{
					PlxPci_DeviceCloseDynamic(&hPlxDevice);

					fnFreePlxDll(g_srDllShared.hPlxDllInstanceLib);
					g_srDllShared.hPlxDllInstanceLib = NULL;
				}

			}

			#ifdef _DEBUG_INFO
			DebugMsg("DLL process detach critical section end");
			#endif
			LeaveCriticalSection(&g_srDllShared.csCBListData);

			break;
			
        case DLL_THREAD_ATTACH:
			#ifdef _DEBUG_INFO
			DebugMsg("DLL_THREAD_ATTACH");
			#endif

			break;
			
        case DLL_THREAD_DETACH:
			
			// obtain the callback list critical section
			// before doing any access/manipulation

			#ifdef _DEBUG_INFO
			DebugMsg("DLL_THREAD_DETACH critical section start");
			#endif

			EnterCriticalSection(&g_srDllShared.csCBListData);
			

			cbNode = cbListHead;
			curThreadId = GetCurrentThreadId();

			while (cbNode)
			{
				if (cbNode->cbThreadId == curThreadId)
				{
					// delete this callback node since it will no longer be needed
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!Deleting a callback node because thread terminated w/o disabling its callback.");
					#endif

					if (cbNode == cbListHead)
						cbListHead = (PCALLBACK_NODE)cbNode->next;
					else
						((PCALLBACK_NODE)cbNode->prev)->next = cbNode->next;	// route around current node in list

					// Adjust the expanded interrupt counts, based on removing
					// the callbacks associated with this thread

					RemoveCallbackCounts(cbNode->cbHooksMask);

					free(cbNode); // remove node
					
					// break because a thread can have only one node in our list
					break;
				}
				else
				{
					#ifdef _DEBUG_INFO
					DebugMsg("DLL_THREAD_DETACH node thread did not match");
					#endif
					cbNode = (PCALLBACK_NODE)cbNode->next;
				}
			}

			#ifdef _DEBUG_INFO
			DebugMsg("DLL_THREAD_DETACH critical section end");
			#endif

			LeaveCriticalSection(&g_srDllShared.csCBListData);

			#ifdef _DEBUG_INFO
			DebugMsg("DLL_THREAD_DETACH done");
			#endif
			break;
        
		default:
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!Unexpected DLLEntryPoint() call with dwReason: %lX",(unsigned long)dwReason);
			#endif
            break;
     }

    return(TRUE);
}


#define MAX_DEVICES_TO_LIST 50 // 50 is some large number we will never exceed
//
//		Name:			InitDLLInstance()
//
//		Description:	On behalf of each application that uses the DLL perform
//						these one-time initializations
//

DWORD InitDLLInstance(void)
{
    //U32  DeviceNum;
    U8   ii;          // use the default int datatype
    U32  NumDevices;
	U32  RegValue;

	BOOLEAN           deviceFound;
	BOOLEAN           bAddDevice, bOpenDevice, bServiceDriver;
	//RETURN_CODE	rc;
	PLX_STATUS        statusFind, statusMisc;
    PLX_DEVICE_KEY    DevKeyToFind;
    PLX_DEVICE_KEY    DevKey_NT[MAX_DEVICES_TO_LIST];
    PLX_PORT_PROP     PortProp;
    PLX_DRIVER_PROP   DriverProp;

	deviceFound = false;

	// Try to find a TCR1000 device

	// The device uses the base PLX codes for device and vendor ID, so
	// we can use the PLX method. However, we must then determine that
	// the subsystem ID is also correct.


	// First we find all of the PLX devices installed
	// that match our initial requirements.
	ii          = 0;
    NumDevices  = 0;

    do
    {
		// Setup to select any device
		memset( &DevKeyToFind, PCI_FIELD_IGNORE, sizeof(PLX_DEVICE_KEY) );

        // Check if device exists
        statusFind = PlxPci_DeviceFindDynamic(&DevKeyToFind, ii);

        if (statusFind != ApiSuccess)
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!PLX device find error: #%d: %d", (int)ii, (int)statusFind);
			#endif
		} else
        {
			#ifdef _DEBUG_INFO
			DebugMsg("PLX find device %d: %d", (int)ii, (int)statusFind);
			#endif
			#ifdef _DEBUG_INFO
			DebugMsg("Port type:%d chip: %04X vendor: %04X device: %04X",
				(int)DevKeyToFind.NTPortType,
				(unsigned int)DevKeyToFind.PlxChip,
				(unsigned int)DevKeyToFind.VendorId,
				(unsigned int)DevKeyToFind.DeviceId);
			#endif

            // Default to add device
            bAddDevice  = TRUE;
			bOpenDevice = FALSE;

            if (DevKeyToFind.NTPortType == PLX_NT_PORT_NONE)
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!PLX find device invalid port == none, should say NT");
				#endif

				// kjf xxx yyy zzz should this be an error?? see PLX NTSample.c
                //bAddDevice = FALSE;
			}

            if (bAddDevice)
            {
                // Verify supported chip type
                if ((DevKeyToFind.PlxChip & 0xFF00) != 0x9000)
                {
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!PLX find device wrong chip: 0x%04X", (unsigned int)DevKeyToFind.PlxChip );
					#endif
                    bAddDevice = FALSE;
                }

                if (DevKeyToFind.VendorId != 0x10B5)	// PLX vendor ID)
                {
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!PLX find device wrong vendor ID: 0x%04X", (unsigned int)DevKeyToFind.VendorId );
					#endif

                    bAddDevice = FALSE;
                }

                if (DevKeyToFind.DeviceId != 0x9030) // PLX device ID for 9030 devices)
                {
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!PLX find device wrong Device ID: 0x%04X", (unsigned int)DevKeyToFind.DeviceId );
					#endif

                    bAddDevice = FALSE;
                }
            }

            if (bAddDevice)
            {
                // Open device to get its properties
                statusMisc = PlxPci_DeviceOpenDynamic(&DevKeyToFind, &hPlxDevice);
				if (statusMisc != ApiSuccess)
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!PLX find device open failed: %d", (int)statusMisc );
					#endif

					bAddDevice = FALSE;
				} else
				{
					#ifdef _DEBUG_INFO
					DebugMsg("PLX find device open successful: %d", (int)statusMisc );
					#endif

					bOpenDevice = TRUE;
				}
            }

            // Verify driver used is NT PnP and not Service driver
            if (bAddDevice)
            {
                statusMisc = PlxPci_DriverPropertiesDynamic(&hPlxDevice, &DriverProp);
				if (statusMisc != ApiSuccess)
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!PLX find device driver properties failed: %d", (int)statusMisc );
					#endif

					bAddDevice = FALSE;
				} else
				{
					#ifdef _DEBUG_INFO
					DebugMsg("isService: %d fullname: %s name: %s version: %08lX",
						(int)DriverProp.bIsServiceDriver,
						DriverProp.FullName,
						DriverProp.Name,
						(unsigned long)DriverProp.Version);
					#endif

					bServiceDriver = DriverProp.bIsServiceDriver;

					if (bServiceDriver)
					{
						#ifdef _DEBUG_ERROR
						DebugMsg("!!!PLX find service driver failed: %d", (int)bServiceDriver );
						#endif

						bAddDevice = FALSE;
					}
				}
            }

            // Get port properties
            if (bAddDevice)
            {
                // Get port properties
                statusMisc = PlxPci_GetPortPropertiesDynamic(&hPlxDevice, &PortProp);
				if (statusMisc != ApiSuccess)
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!PLX find device driver properties failed: %d", (int)statusMisc );
					#endif

					bAddDevice = FALSE;
				}
            }

			if (bAddDevice)
			{
				#ifdef _DEBUG_INFO
				DebugMsg("port number: %d port type: %d",
					(unsigned int)PortProp.PortNumber,
					(unsigned int)PortProp.PortType);
				#endif

				// Read the Subsystem Device/Vendor ID
				RegValue = PlxPci_PciRegisterReadFastDynamic(&hPlxDevice, CFG_SUB_VENDOR_ID, &statusMisc);

				// Determine that the subsystem id matches the supported devices
				// The only currently supported device is the TCR1000

				// TCR1000 - 0x40c05743
				if (statusMisc == ApiSuccess)
				{
					if (RegValue == 0x40c03525)
					{
						#ifdef _DEBUG_INFO
						DebugMsg("PLX valid device found");
						#endif

						// this is our device, exit the loop
						deviceFound = true;
						devInfo.devType = TC_PRODUCT_TCR1000_X64_REV_A;
						break;
					} else
					{
						#ifdef _DEBUG_ERROR
						DebugMsg("!!!PLX find device invalid sub vendor id: %lX", (unsigned long)RegValue );
						#endif
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!PLX find device read register failed: %d", (int)statusMisc );
					#endif
				}
			}

            // Not the device we want, Close device
			if (bOpenDevice)
				PlxPci_DeviceCloseDynamic(&hPlxDevice);

            if (bAddDevice)
            {
                // Copy device key info
                DevKey_NT[NumDevices] = DevKeyToFind;

				#ifdef _DEBUG_INFO
                DebugMsg("\t\t  %2d. %04x Port %d [b:%02x s:%02x] (%s-side)\n",
                            (int)(NumDevices + 1),
							(unsigned int)DevKeyToFind.PlxChip,
							(int)PortProp.PortNumber,
                            (unsigned int)DevKeyToFind.bus,
							(unsigned int)DevKeyToFind.slot,
                            (DevKeyToFind.NTPortType == PLX_NT_PORT_LINK) ? "Link" : "Virtual");
				#endif

                // Increment device count
                NumDevices++;
            }
        }

        // Go to next devices
        ii++;
    }
    while ((statusFind == ApiSuccess) && (NumDevices < MAX_DEVICES_TO_LIST));

    if (!deviceFound)
	{
		// TCR1000 not found
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!InitDLLInstance: TCR1000 not found");
		#endif
        return DLL_INIT_NOKMD;
	}
	

	// We have a device, now save the handle to the
	// device for future calls.
	plxIoBuffer.Key = DevKeyToFind;

	// Create the Mutexes necessary for controlling access to the shared data.
	// Since the Mutex could be initially created by a privileged process
	// (like TCRSync-1000) that won't allow other processes to access the Mutex,
	// we need SECURITY_ATTRIBUTES and a SECURITY_DESCRIPTOR that will allow
	// all process access to the Mutex. Allocating the SECURITY_DESCRIPTOR assures
	// that the structure will be on the correct word boundary for the support
	// functions to work.

    SECURITY_ATTRIBUTES secAttr;
    PSECURITY_DESCRIPTOR pSecDesc = malloc(sizeof(SECURITY_DESCRIPTOR));
    if (pSecDesc == 0)
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!InitDLLInstance: malloc failed");
		#endif

		PlxPci_DeviceCloseDynamic(&hPlxDevice);
		return DLL_INIT_UNEXPECTED;
	}
    
	secAttr.nLength = sizeof(secAttr);
	secAttr.lpSecurityDescriptor = (LPVOID) pSecDesc;
	secAttr.bInheritHandle = FALSE; // child processes do not inherit the handle

	InitializeSecurityDescriptor(pSecDesc, SECURITY_DESCRIPTOR_REVISION);

	// Set the "discretionary ACL (access control list)" to provide all access to
	// the object
	SetSecurityDescriptorDacl(pSecDesc, TRUE, NULL, TRUE);

	// Create the mutex that controls access to the CB counts

	g_srDllShared.hCBCountMutex = CreateMutex(&secAttr, FALSE, "Mutex_MC_CBCount");
	if (g_srDllShared.hCBCountMutex == NULL)
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!InitDLLInstance: create mutex g_srDllShared.hCBCountMutex failed");
		#endif

		PlxPci_DeviceCloseDynamic(&hPlxDevice);
		free(pSecDesc);
		return DLL_INIT_UNEXPECTED;
	}

	// Create the mutex that controls to the LastInt data

	g_srDllShared.hLastIntMutex = CreateMutex(&secAttr, FALSE, "Mutex_MC_LastInt");
	if (g_srDllShared.hLastIntMutex == NULL)
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!InitDLLInstance: create mutex g_srDllShared.hLastIntMutex failed");
		#endif

		PlxPci_DeviceCloseDynamic(&hPlxDevice);
		free(pSecDesc);
		return DLL_INIT_UNEXPECTED;
	}

	free(pSecDesc); // no longer needed

	// Create the critical section object that controls the access
	// to the callback data.

	InitializeCriticalSection(&g_srDllShared.csCBListData);

	// Create the AsyncEventMonitor for the PCI interrupt

	if (CreateThread(NULL, 0, AsyncEventMonitor, NULL, 0,
					 &asyncMonitorId) == NULL)
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!InitDLLInstance: create AsyncEventMonitor failed");
		#endif

		PlxPci_DeviceCloseDynamic(&hPlxDevice);
		return DLL_INIT_UNEXPECTED;
	}

	#ifdef _DEBUG_INFO
	DebugMsg("InitDLLInstance completed successfully");
	#endif

	return DLL_INIT_OK;
}

//
//		Name:			AcquireApiMutex()
//
//		Description:	Waits for the API mutex. If the wait fails, just continue.
//                      It is possible that if a process has acquired the mutex, but, before the process
//                      can free the mutex it is deleted, the mutex would forever be allocated. Instead,
//                      we continue after the timeout.
// 
//
#define MUTEX_RETRY_COUNT 30
#define MUTEX_RETRY_MSEC  10
void AcquireApiMutex(void)
{
	LONG lPriorValue;

	int ii;

	// simulate a 300 msec timeout
	for (ii=0; ii<MUTEX_RETRY_COUNT; ii++)
	{
		if (!g_srDllShared.ldllApiMutex)
		{
			lPriorValue = ::InterlockedExchange(&g_srDllShared.ldllApiMutex, 1);

			if (!lPriorValue)
			{
				// It was a zero and we set it high, we can continue
				break;
			}

			// Another thread beat us to setting it to a one.
			// Wait for that thread to zero the value

		}

		Sleep(MUTEX_RETRY_MSEC);
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!acquiring DLL mutex retry");
		#endif
	}

	if (ii >= MUTEX_RETRY_COUNT)
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Failed acquiring DLL mutex");
		#endif
	}
}


//
//		Name:			ReleaseApiMutex()
//
//		Description:	Releases the API mutex.
//
//

void ReleaseApiMutex(void)
{
	//HANDLE hProcessID;
	//DWORD  dwPriority;

	//hProcessID = GetCurrentProcess();
	//dwPriority = GetPriorityClass(hProcessID);
	//SetPriorityClass(hProcessID,REALTIME_PRIORITY_CLASS);

	::InterlockedExchange(&g_srDllShared.ldllApiMutex, 0);

	//SetPriorityClass(hProcessID,dwPriority);

}





//
//		Name:			PopulateTimeFrame()
//
//		Description:	Build a TIMEFRAME structure from the data returned by TCR kernel-mode
//						driver.
//
//		Returns:		0 if data read OK
//						DLL_TC_STATUS_DRVFAIL if failure 
//

DWORD PopulateTimeFrame(TIMEFRAME *ptimeFrameReturned, LARGE_INTEGER liPerfCount)
{
	int ii,jj;

	RETURN_CODE rc;
	DWORD       dwStatus;

	TIMEFRAME timeFrameRead;
	
	// First, read the interrupt performance count (the time of the previous interrupt)
	// if there was no value passed in.

	LARGE_INTEGER liPerfCount2;

	if (liPerfCount.QuadPart == 0) {

		// This is a call into the PLX DLL that we added
		rc = McIntrCountGetDynamic(&hPlxDevice, &liPerfCount);

		if (rc != ApiSuccess)
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!PopulateTimeFrame: get count #1 failed: %d",(int)rc);
			#endif

			return DLL_TC_STATUS_DRVFAIL;	// fail
		}
	}

	// Ensure that we have a consistent set of data. This is done with two checks.
	//
	// 1) The last interrupt time is checked both before and after the read.
	//	  If there is no change, then no interrupt happened.
	// 2) Verify the checksum of the data buffer.
	//
	// We will try three times to achieve this, though in normal operation
	// only a single retry should be necessary.

	bool tcValid = false;
	bool bRetry  = false;
	for (ii=0; (ii<5) && (!tcValid); ii++) {

		dwStatus = ReadTimeFrameFromPlx(&timeFrameRead);

		if (dwStatus != DLL_TC_STATUS_NORMAL)
		{
			return dwStatus;
		}

		// Validate the timecode data against the checksum

		U8 tmpSum = timeFrameRead.tcBuffer[0];
		for (jj=1; jj<=7; jj++)
			tmpSum += timeFrameRead.tcBuffer[jj];

		if (bRetry)
		{
			// print out the data received on the retry
			bRetry = false;
			#ifdef _DEBUG_INFO2
			DebugMsg("retry: %02X %02X %02X %02X     %02X %02X %02X %02X",
				timeFrameRead.tcBuffer[0], timeFrameRead.tcBuffer[1], timeFrameRead.tcBuffer[2], timeFrameRead.tcBuffer[3],
				timeFrameRead.tcBuffer[4], timeFrameRead.tcBuffer[5], timeFrameRead.tcBuffer[6], timeFrameRead.tcBuffer[7]);
			#endif
		}

		if (tmpSum != timeFrameRead.tcChecksum)
		{
			#ifdef _DEBUG_INFO2
			DebugMsg("!!!PopulateTimeFrame: tcChecksum failed verify calc:%02X rx:%02X",tmpSum,timeFrameRead.tcChecksum);
			DebugMsg("       %02X %02X %02X %02X     %02X %02X %02X %02X",
				timeFrameRead.tcBuffer[0], timeFrameRead.tcBuffer[1], timeFrameRead.tcBuffer[2], timeFrameRead.tcBuffer[3],
				timeFrameRead.tcBuffer[4], timeFrameRead.tcBuffer[5], timeFrameRead.tcBuffer[6], timeFrameRead.tcBuffer[7]);
			#endif
			bRetry = TRUE;
			continue;	// retry
		}

		// reread the interrupt count to check for consistency (no new interrupt while reading data)

		// This is a call into the PLX DLL that we added
		rc = McIntrCountGetDynamic(&hPlxDevice, &liPerfCount2);

		if (rc != ApiSuccess)
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!PopulateTimeFrame: get count #2 failed: %d",(int)rc);
			#endif
			return DLL_TC_STATUS_DRVFAIL;	// fail
		}

		if (liPerfCount.QuadPart != liPerfCount2.QuadPart) {
			#ifdef _DEBUG_INFO2
			DebugMsg("!!!PopulateTimeFrame: get count diff, retry: %ld::%ld",(long)liPerfCount.QuadPart, (long)liPerfCount2.QuadPart);
			#endif
			// new interrupt, retry
			liPerfCount.QuadPart = liPerfCount2.QuadPart;
		}
		else {
			tcValid = true;
			timeFrameRead.intPerfCount.QuadPart = liPerfCount2.QuadPart;
		}
	}

	if (tcValid && (ii<5))
	{
		ptimeFrameReturned->intPerfCount = timeFrameRead.intPerfCount;
		ptimeFrameReturned->tcChecksum   = timeFrameRead.tcChecksum;
		ptimeFrameReturned->tcStatus     = timeFrameRead.tcStatus;
		memcpy(ptimeFrameReturned->tcBuffer, timeFrameRead.tcBuffer, sizeof(ptimeFrameReturned->tcBuffer));
		return DLL_TC_STATUS_NORMAL;	// success
	}
	else
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!PopulateTimeFrame failed reading a valid time code");
		#endif

		return DLL_TC_STATUS_INVALID;	// failure
	}
}

//
//		Name:			ReadTimeFrameFromPlx()
//
//		Description:	Reads the TIMEFRAME with retries if the data changed.
//
//		Returns:		TRUE if time code successfully processed
//						FALSE if no time code present 
//

DWORD ReadTimeFrameFromPlx(TIMEFRAME *timeFrameToReturn)
{
	RETURN_CODE rc;
	int  ii;
	bool bRetry;
	TIMEFRAME timeFrameRead[2];

	// Read the time frame up to 5 times trying to get two copies that are the same.
	memset(timeFrameRead, 0, sizeof(timeFrameRead));

	for (ii=0; ii<5; ii++) {
		// Read the status, buffer, and checksum off of the card
		//rc = TCRRead(TCR_READ_TC_STATUS, &(t->tcStatus), 1, BitSize8);
		rc = PlxPci_PciBarSpaceReadDynamic(&hPlxDevice, 2, TCR_READ_TC_STATUS, &(timeFrameRead[0].tcStatus), 1, BitSize8, FALSE);

		if ( rc != ApiSuccess)
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!PopulateTimeFrame: get tcStatus failed: %d",(int)rc);
			#endif
			return DLL_TC_STATUS_DRVFAIL;	// fail
		}

		//rc = TCRRead(TCR_READ_DATA_BUFFER, t->tcBuffer, 8, BitSize32);
		rc = PlxPci_PciBarSpaceReadDynamic(&hPlxDevice, 2, TCR_READ_DATA_BUFFER, timeFrameRead[0].tcBuffer, 8, BitSize32, FALSE);
		if (rc != ApiSuccess)
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!PopulateTimeFrame: get tcBuffer failed: %d",(int)rc);
			#endif
			return DLL_TC_STATUS_DRVFAIL;	// fail
		}

		//rc = TCRRead(TCR_READ_DATA_CHKSUM, &(t->tcChecksum), 1, BitSize8);
		rc = PlxPci_PciBarSpaceReadDynamic(&hPlxDevice, 2, TCR_READ_DATA_CHKSUM, &(timeFrameRead[0].tcChecksum), 1, BitSize8, FALSE);
		if ( rc != ApiSuccess)
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!PopulateTimeFrame: get tcChecksum failed: %d",(int)rc);
			#endif
			return DLL_TC_STATUS_DRVFAIL;	// fail
		}

		// Read the status, buffer, and checksum off of the card
		//rc = TCRRead(TCR_READ_TC_STATUS, &(t->tcStatus), 1, BitSize8);
		rc = PlxPci_PciBarSpaceReadDynamic(&hPlxDevice, 2, TCR_READ_TC_STATUS, &(timeFrameRead[1].tcStatus), 1, BitSize8, FALSE);

		if ( rc != ApiSuccess)
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!PopulateTimeFrame: get tcStatus failed: %d",(int)rc);
			#endif
			return DLL_TC_STATUS_DRVFAIL;	// fail
		}

		//rc = TCRRead(TCR_READ_DATA_BUFFER, t->tcBuffer, 8, BitSize32);
		rc = PlxPci_PciBarSpaceReadDynamic(&hPlxDevice, 2, TCR_READ_DATA_BUFFER, timeFrameRead[1].tcBuffer, 8, BitSize32, FALSE);
		if (rc != ApiSuccess)
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!PopulateTimeFrame: get tcBuffer failed: %d",(int)rc);
			#endif
			return DLL_TC_STATUS_DRVFAIL;	// fail
		}

		//rc = TCRRead(TCR_READ_DATA_CHKSUM, &(t->tcChecksum), 1, BitSize8);
		rc = PlxPci_PciBarSpaceReadDynamic(&hPlxDevice, 2, TCR_READ_DATA_CHKSUM, &(timeFrameRead[1].tcChecksum), 1, BitSize8, FALSE);
		if ( rc != ApiSuccess)
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!PopulateTimeFrame: get tcChecksum failed: %d",(int)rc);
			#endif
			return DLL_TC_STATUS_DRVFAIL;	// fail
		}

		// we have two copies, if they are not the same, try again
		bRetry = false;
		if (timeFrameRead[0].tcStatus != timeFrameRead[1].tcStatus)
		{
			#ifdef _DEBUG_INFO2
			DebugMsg("!!!PopulateTimeFrame retry tcStatus change: %d %d",(int)timeFrameRead[0].tcStatus, timeFrameRead[1].tcStatus);
			#endif
			bRetry = true;
		}

		if (memcmp(timeFrameRead[0].tcBuffer, timeFrameRead[1].tcBuffer, sizeof(timeFrameRead[0].tcBuffer)))
		{
			#ifdef _DEBUG_INFO2
			DebugMsg("!!!PopulateTimeFrame retry tcBuffer change");
			#endif
			bRetry = true;
		}

		if (timeFrameRead[0].tcChecksum != timeFrameRead[1].tcChecksum)
		{
			#ifdef _DEBUG_INFO2
			DebugMsg("!!!PopulateTimeFrame retry tcChecksum change: %d %d",(int)timeFrameRead[0].tcChecksum, timeFrameRead[1].tcChecksum);
			#endif
			bRetry = true;
		}

		if (!bRetry)
			break;
	}

	if (ii >= 5)
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!ReadTimeFrameFromPlx exceeded retries max: %d",(int)ii);
		#endif

		return DLL_TC_STATUS_DRVFAIL;
	}

	timeFrameToReturn->tcChecksum = timeFrameRead[0].tcChecksum;
	timeFrameToReturn->tcStatus   = timeFrameRead[0].tcStatus;
	memcpy(timeFrameToReturn->tcBuffer, timeFrameRead[0].tcBuffer, sizeof(timeFrameToReturn->tcBuffer));
	return DLL_TC_STATUS_NORMAL;
}

//
//		Name:			ProcessTimeFrame()
//
//		Description:	Build a TCRTIME structure from the t structure returned by TCR kernel-mode
//						driver.
//
//		Returns:		TRUE if time code successfully processed
//						FALSE if no time code present 
//

BOOL ProcessTimeFrame(TIMEFRAME *timeFrame, TCRTIME *tcrt)
{
	WORD	tcnt;
	BOOL	tmpflag;

	// Decode the time value, based on if we have SMPTE or IRIG

	switch (timeFrame->tcStatus) {
		case 0:		// 30 fps
		case 1:		// 30 fps drop frame
		case 2:		// 25 fps
		case 3:		// 24 fps
			// SMPTE

			tcrt->dwTimeCodeType = TC_TYPE_SMPTE;

			// Populate SMPTE structure from tcBuffer
			//MSF fixed bits that should not be used for time
			tcrt->rawTc.smpte.fu =  timeFrame->tcBuffer[0] & 0x0F;
			tcrt->rawTc.smpte.ft = (timeFrame->tcBuffer[0] >> 4) & 0x03;  //0x0F;
			tcrt->rawTc.smpte.su =  timeFrame->tcBuffer[1] & 0x0F;
			tcrt->rawTc.smpte.st = (timeFrame->tcBuffer[1] >> 4) & 0x07;  //0x0F;
			tcrt->rawTc.smpte.mu =  timeFrame->tcBuffer[2] & 0x0F;
			tcrt->rawTc.smpte.mt = (timeFrame->tcBuffer[2] >> 4) & 0x07;  //0x0F;
			tcrt->rawTc.smpte.hu =  timeFrame->tcBuffer[3] & 0x0F;
			tcrt->rawTc.smpte.ht = (timeFrame->tcBuffer[3] >> 4) & 0x03;  //0x0F;
			tcrt->rawTc.smpte.u1 =  timeFrame->tcBuffer[4] & 0x0F;
			tcrt->rawTc.smpte.u2 = (timeFrame->tcBuffer[4] >> 4) & 0x0F;
			tcrt->rawTc.smpte.u3 =  timeFrame->tcBuffer[5] & 0x0F;
			tcrt->rawTc.smpte.u4 = (timeFrame->tcBuffer[5] >> 4) & 0x0F;
			tcrt->rawTc.smpte.u5 =  timeFrame->tcBuffer[6] & 0x0F;
			tcrt->rawTc.smpte.u6 = (timeFrame->tcBuffer[6] >> 4) & 0x0F;
			tcrt->rawTc.smpte.u7 =  timeFrame->tcBuffer[7] & 0x0F;
			tcrt->rawTc.smpte.u8 = (timeFrame->tcBuffer[7] >> 4) & 0x0F;

			// Populate the time data (raw data is BCD)
			tcrt->wHour	  = (tcrt->rawTc.smpte.ht * 10) + tcrt->rawTc.smpte.hu;
			tcrt->wMinute = (tcrt->rawTc.smpte.mt * 10) + tcrt->rawTc.smpte.mu;
			tcrt->wSecond = (tcrt->rawTc.smpte.st * 10) + tcrt->rawTc.smpte.su;
			{
				WORD frameNumWord = (WORD)((tcrt->rawTc.smpte.ft * 10) + tcrt->rawTc.smpte.fu) * 1000;
				if (timeFrame->tcStatus == 3) {
					tcrt->wMilliseconds = frameNumWord / 24;
				} else if (timeFrame->tcStatus == 2) {
					tcrt->wMilliseconds = frameNumWord / 25;
				} else {
					tcrt->wMilliseconds = frameNumWord / 30;
				}
			}

			// Decode Leitch-format date from user groups
			// Note: the value of these members will be undetermined if
			// the Leitch-format date isn't present
			// Day tens are two LSB's of group 4. Day units are group 2
			tcrt->wDay = ((tcrt->rawTc.smpte.u4 & 0x03) * 10) + tcrt->rawTc.smpte.u2;

			// Month ten is 3rd LSB of group 4. Month units are group 3 
			tcrt->wMonth = (((tcrt->rawTc.smpte.u4 >> 2) & 0x01) * 10) + tcrt->rawTc.smpte.u3;

			// Year tens is group 8. Year units is group 6.
			// This code has a Y2.1K problem..
			tcrt->wYear	= 2000 + ((tcrt->rawTc.smpte.u8) * 10) + tcrt->rawTc.smpte.u6;

			// Pass on the interrupt value

			tcrt->intPerfCount.QuadPart = timeFrame->intPerfCount.QuadPart;

			// decode lock status
			// note: the value of this member will be undetermined if
			// the time code source is not a Masterclock device with SMPTE, Leitch date encoding, with Masterclock proprietary control bits

			devInfo.gpsStatus = (tcrt->rawTc.smpte.u5 << 4) | tcrt->rawTc.smpte.u1;

			// however, we're only supporting freewheeling right now

			devInfo.gpsStatus &= 0x01;

			// Data is valid

			tcrt->valid = TRUE;

			// Perform a rough validity check of the date/time
			if ((tcrt->wMonth < 1) || (tcrt->wMonth > 12))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid month: %d", (int)tcrt->wMonth);
				#endif
				tcrt->valid = FALSE;
			}

			if ((tcrt->wDay < 1) || (tcrt->wDay > 31))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid date: %d", (int)tcrt->wDay);
				#endif
				tcrt->valid = FALSE;
			}

			if ((tcrt->wYear < 1900) || (tcrt->wYear > 2100))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid year: %d", (int)tcrt->wYear);
				#endif
				tcrt->valid = FALSE;
			}

			if ((tcrt->wHour < 0) || (tcrt->wHour > 60))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid hour: %d", (int)tcrt->wHour);
				#endif
				tcrt->valid = FALSE;
			}

			if ((tcrt->wMinute < 0) || (tcrt->wMinute > 60))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid minute: %d", (int)tcrt->wMinute);
				#endif
				tcrt->valid = FALSE;
			}

			if ((tcrt->wSecond < 0) || (tcrt->wSecond > 60))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid Second: %d", (int)tcrt->wSecond);
				#endif
				tcrt->valid = FALSE;
			}

			if ((tcrt->wMilliseconds < 0) || (tcrt->wMilliseconds > 999))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid msec: %d", (int)tcrt->wMilliseconds);
				#endif
				tcrt->valid = FALSE;
			}

			if (tcrt->valid == TRUE)
				return TRUE;
			else
				return FALSE;

			break;
			break;

		case 4:
		case 5:
			// IRIG

			if (timeFrame->tcStatus == 4)
				tcrt->dwTimeCodeType = TC_TYPE_IRIG_B1;
			else
				tcrt->dwTimeCodeType = TC_TYPE_IRIG_B;

			// Populate IRIG structure from tcBuffer
			tcrt->rawTc.irig.su  =  timeFrame->tcBuffer[0] & 0x0F;
			tcrt->rawTc.irig.st  = (timeFrame->tcBuffer[0] >> 4) & 0x0F;
			tcrt->rawTc.irig.mu  =  timeFrame->tcBuffer[1] & 0x0F;
			tcrt->rawTc.irig.mt  = (timeFrame->tcBuffer[1] >> 4) & 0x0F;
			tcrt->rawTc.irig.hu  =  timeFrame->tcBuffer[2] & 0x0F;
			tcrt->rawTc.irig.ht  = (timeFrame->tcBuffer[2] >> 4) & 0x0F;
			tcrt->rawTc.irig.du  =  timeFrame->tcBuffer[3] & 0x0F;
			tcrt->rawTc.irig.dt  = (timeFrame->tcBuffer[3] >> 4) & 0x0F;
			tcrt->rawTc.irig.dh  =  timeFrame->tcBuffer[4] & 0x0F;
			tcrt->rawTc.irig.cf1 = (timeFrame->tcBuffer[4] >> 4) & 0x0F;
			tcrt->rawTc.irig.cf2 =  timeFrame->tcBuffer[5] & 0x0F;
			tcrt->rawTc.irig.cf3 = (timeFrame->tcBuffer[5] >> 4) & 0x0F;
			tcrt->rawTc.irig.cf4 =  timeFrame->tcBuffer[6] & 0x0F;
			tcrt->rawTc.irig.cf5 = (timeFrame->tcBuffer[6] >> 4) & 0x0F;
			tcrt->rawTc.irig.cf6 =  timeFrame->tcBuffer[7] & 0x0F;
			tcrt->rawTc.irig.cf7 = (timeFrame->tcBuffer[7] >> 4) & 0x0F;

			// Populate the time data (raw data is BCD)
			tcrt->wHour	  = (tcrt->rawTc.irig.ht * 10) + tcrt->rawTc.irig.hu;
			tcrt->wMinute = (tcrt->rawTc.irig.mt * 10) + tcrt->rawTc.irig.mu;
			tcrt->wSecond = (tcrt->rawTc.irig.st * 10) + tcrt->rawTc.irig.su;
			tcrt->wMilliseconds = 0;	// Irig cannot do subsecond timing.

			// decode IEEE1344-format date from control functions
			// note: the value of these members will be undetermined if
			// the IEEE1344 encoding isn't present

			tcrt->wYear = tcrt->rawTc.irig.cf1 & 0x0f;
			tcrt->wYear += (((tcrt->rawTc.irig.cf2 >> 1) | ((tcrt->rawTc.irig.cf3 & 0x01) << 3)) * 10);

			// Adjust to the 21st century. This code has a Y2.1K problem.

			tcrt->wYear += 2000;

			// Compute the day-of-year, and then get the month and day-of-month
			// (depends on if this is a leap year.

			tcrt->wDay = (tcrt->rawTc.irig.dh * 100) +
				         (tcrt->rawTc.irig.dt * 10) +
						  tcrt->rawTc.irig.du;

			tmpflag = !(tcrt->wYear & 0x03);

			// determine month and day
		
			if (tmpflag)
			{
				for(tcnt = 1; tcnt < 13; tcnt++)
				{
					if (_daysMonthElapsedLeapYear[tcnt] >= tcrt->wDay)
					{
						tcrt->wMonth = tcnt;
						tcrt->wDay   = tcrt->wDay - _daysMonthElapsedLeapYear[tcnt-1];
						break;
					}
				}
			}
			else
			{
				for(tcnt = 1; tcnt < 13; tcnt++)
				{
					if (_daysMonthElapsedNonLeapYear[tcnt] >= tcrt->wDay)
					{
						tcrt->wMonth = tcnt;
						tcrt->wDay   = tcrt->wDay - _daysMonthElapsedNonLeapYear[tcnt-1];
						break;
					}
				}
			}	

			// Pass on the interrupt value

			tcrt->intPerfCount.QuadPart = timeFrame->intPerfCount.QuadPart;

			tcrt->valid = TRUE;

			// Perform a rough validity check of the date/time
			if ((tcrt->wMonth < 1) || (tcrt->wMonth > 12))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid month: %d", (int)tcrt->wMonth);
				#endif
				tcrt->valid = FALSE;
			}

			if ((tcrt->wDay < 1) || (tcrt->wDay > 31))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid date: %d", (int)tcrt->wDay);
				#endif
				tcrt->valid = FALSE;
			}

			if ((tcrt->wYear < 1900) || (tcrt->wYear > 2100))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid year: %d", (int)tcrt->wYear);
				#endif
				tcrt->valid = FALSE;
			}

			if ((tcrt->wHour < 0) || (tcrt->wHour > 60))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid hour: %d", (int)tcrt->wHour);
				#endif
				tcrt->valid = FALSE;
			}

			if ((tcrt->wMinute < 0) || (tcrt->wMinute > 60))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid minute: %d", (int)tcrt->wMinute);
				#endif
				tcrt->valid = FALSE;
			}

			if ((tcrt->wSecond < 0) || (tcrt->wSecond > 60))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid Second: %d", (int)tcrt->wSecond);
				#endif
				tcrt->valid = FALSE;
			}

			if ((tcrt->wMilliseconds < 0) || (tcrt->wMilliseconds > 999))
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!Invalid msec: %d", (int)tcrt->wMilliseconds);
				#endif
				tcrt->valid = FALSE;
			}

			if (tcrt->valid == TRUE)
				return TRUE;
			else
				return FALSE;
			break;

		default:
			return FALSE;	// TC not present (either lost or never there)
	}
	
	return FALSE;	// error	
}


//
//		Name:			AsyncEventMonitor()
//
//		Description:	Monitors TCR driver for asynchronous notifications, process them, and calls
//						callback functions with required data.
//
//		Returns:		When hMonitorExit is signalled.
//

DWORD WINAPI AsyncEventMonitor(LPVOID lpParam)
{
    PLX_INTERRUPT   PlxInterrupt;
    PLX_NOTIFY_OBJECT NotifyObject;

	PLX_STATUS		rc;

	TIMEFRAME		interruptTimeframe;
	TCRTIME			lTime;
	DWORD			intDataRc;
	PCALLBACK_NODE	cbNode;

	#ifdef _DEBUG_INFO
	DebugMsg("PLX Event Monitor() initialized.");
	#endif

	bool waitActive = true;
	while (waitActive)
	{
		// Attach event to Plx interrupt
		memset(&PlxInterrupt, 0, sizeof(PLX_INTERRUPT));
		PlxInterrupt.PciMain     = 1;

		rc = PlxPci_InterruptEnableDynamic(&hPlxDevice, &PlxInterrupt);
		if (rc != ApiSuccess)
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!PLX Event Monitor: int enable failed: %d",(int)rc);
			#endif
			waitActive = false;
			break;
		}

		memset(&PlxInterrupt, 0, sizeof(PLX_INTERRUPT));
		PlxInterrupt.LocalToPci  = 1;
		rc = PlxPci_NotificationRegisterForDynamic(&hPlxDevice, &PlxInterrupt, &NotifyObject);
		if (rc != ApiSuccess)
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!PLX Event Monitor: register int failed: %d",(int)rc);
			#endif
			waitActive = false;
			break;
		}

		// Wait for one of the events to happen
		// Wait for interrupt event
		rc = PlxPci_NotificationWaitDynamic(&hPlxDevice, &NotifyObject, 10 * 1000);   // wait up to ten seconds

		switch (rc) {
			case ApiSuccess:

				// Get the interrupt status and data
				// Data is only populated if the interrupt is a timing interrupt

				UCHAR interruptReason;

				intDataRc = GetInterruptData(&interruptReason, &interruptTimeframe);

				if (intDataRc)
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!PLX notification error INT=%d, rc=%d", interruptReason, intDataRc);
					#endif
					break;	// fail code
				}

				#ifdef _DEBUG_INFO
				DebugMsg("PLX notification INT=%d, rc=%d", interruptReason, intDataRc);
				#endif

				// If the interrupt is a timing interrupt, process the interrupt data

				if (interruptReason & ASYNC_TIMING_INTS) {

					// Data read, now process it into the correct format

					ProcessTimeFrame(&interruptTimeframe, &lTime);
				}

				// asynchronous time completed

				// scan callback list for threads which have requested callbacks for this asynchronous
				// notification

				// protect ourselves from accessing the callback list while it is being changed by 
				// obtaining the mutex

				#ifdef _DEBUG_INFO
				DebugMsg("PLX event critical section start");
				#endif
				EnterCriticalSection(&g_srDllShared.csCBListData);

				cbNode = cbListHead;

				while (cbNode)
				{
					// Check for each of the 6 different kinds of interrupts,
					// and call any attached functions

					if (interruptReason & ASYNC_PERSEC &&
						cbNode->cbHooksMask & ASYNC_PERSEC)
					{

						CopyMemory(&cbNode->timeData,&lTime,sizeof(TCRTIME));

						#ifdef _DEBUG_INFO2
						DebugMsg("PLX callback ASYNC_PERSEC %d:%d:%d %d/%d/%d",
							(int)lTime.wHour,  (int)lTime.wMinute, (int)lTime.wSecond,
							(int)lTime.wMonth, (int)lTime.wDay,    (int)lTime.wYear);
						#endif	

						(*(cbNode->cbFunction))(ASYNC_PERSEC,&cbNode->timeData,sizeof(TCRTIME));

						#ifdef _DEBUG_INFO
						DebugMsg("PLX callback ASYNC_PERSEC done");
						#endif					
					}

					if (interruptReason & ASYNC_TIMESET &&
						cbNode->cbHooksMask & ASYNC_TIMESET)
					{

						CopyMemory(&cbNode->timeData,&lTime,sizeof(TCRTIME));

						#ifdef _DEBUG_INFO
						DebugMsg("PLX callback ASYNC_TIMESET %d:%d:%d %d/%d/%d",
							(int)lTime.wHour,  (int)lTime.wMinute, (int)lTime.wSecond,
							(int)lTime.wMonth, (int)lTime.wDay,    (int)lTime.wYear);
						#endif	

						(*(cbNode->cbFunction))(ASYNC_TIMESET,&cbNode->timeData,sizeof(TCRTIME));
					}

					if (interruptReason & ASYNC_STARTTC &&
						cbNode->cbHooksMask & ASYNC_STARTTC)
					{
						CopyMemory(&cbNode->timeData,&lTime,sizeof(TCRTIME));

						#ifdef _DEBUG_INFO
						DebugMsg("PLX callback ASYNC_STARTTC %d:%d:%d %d/%d/%d",
							(int)lTime.wHour,  (int)lTime.wMinute, (int)lTime.wSecond,
							(int)lTime.wMonth, (int)lTime.wDay,    (int)lTime.wYear);
						#endif	

						(*(cbNode->cbFunction))(ASYNC_STARTTC,&cbNode->timeData,sizeof(TCRTIME));

					}

					if (interruptReason & ASYNC_ENDTC &&
						cbNode->cbHooksMask & ASYNC_ENDTC)
					{
						CopyMemory(&cbNode->timeData,&lTime,sizeof(TCRTIME));

						#ifdef _DEBUG_INFO
						DebugMsg("TC ASYNC_ENDTC %d:%d:%d %d/%d/%d",
							(int)lTime.wHour,  (int)lTime.wMinute, (int)lTime.wSecond,
							(int)lTime.wMonth, (int)lTime.wDay,    (int)lTime.wYear);
						#endif	

						(*(cbNode->cbFunction))(ASYNC_ENDTC,&cbNode->timeData,sizeof(TCRTIME));
					}

					if (interruptReason & ASYNC_TIMELOSS &&
						cbNode->cbHooksMask & ASYNC_TIMELOSS)
					{
						#ifdef _DEBUG_INFO
						DebugMsg("TC ASYNC_TIMELOSS");
						#endif	

						(*(cbNode->cbFunction))(ASYNC_TIMELOSS,NULL,0);

					}

					if (interruptReason & ASYNC_TIMEGAIN &&
						cbNode->cbHooksMask & ASYNC_TIMEGAIN)
					{
						#ifdef _DEBUG_INFO
						DebugMsg("TC ASYNC_TIMEGAIL");
						#endif

						(*(cbNode->cbFunction))(ASYNC_TIMEGAIN,NULL,0);	
					}

					cbNode = (PCALLBACK_NODE)cbNode->next;
				}

				#ifdef _DEBUG_INFO
				DebugMsg("PLX event critical section end");
				#endif
				LeaveCriticalSection(&g_srDllShared.csCBListData);

				break;

			case ApiWaitTimeout:
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!PLX Event Monitor: Timeout waiting for Int Event\n");
				#endif
				break;

			case ApiWaitCanceled:
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!PLX Event Monitor: Interrupt event canceled\n");
				#endif
				break;

			default:
				// exit the wait loop
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!PLX Event Monitor: Interrupt event unknown code:%d\n",rc);
				#endif

				waitActive = false;	
				break;
		}

		rc = PlxPci_NotificationCancelDynamic(&hPlxDevice,&NotifyObject);
		if (rc != ApiSuccess)
		{
			#ifdef _DEBUG_ERROR
			DebugMsg("!!!PLX Notif cancel failed: %d\n",rc);
			#endif
		}

		#ifdef _DEBUG_INFO
		DebugMsg("PLX Event Monitor complete");
		#endif
	}

	#ifdef _DEBUG_INFO
	DebugMsg("AsyncEventMonitor() is terminating.");
	#endif

	return(0);
}


//
//		Name:			GetTCTypeFromTCStatus()
//
//		Description:	Converts the time code status from the device to a time code type.
//

DWORD GetTCTypeFromTCStatus(U8 tcStatus)
{
		// Set the time code type
	switch (tcStatus & 0x7F) {
	case 0:
	case 1:
	case 2:
	case 3:
		return TC_TYPE_SMPTE;
	case 4:
		return TC_TYPE_IRIG_B1;
	case 5:
		return TC_TYPE_IRIG_B;
	default:
		return TC_TYPE_UNDEFINED;
	}
}

//
//		Name:			AddCallbackCounts
//
//		Description:	Adjusts the callback count based on the ending of a thread.
//
//		Returns 0 is success, or ASYNC_MASK_ERROR if there is an error.
//

DWORD AddCallbackCounts(DWORD cbToAdd)
{
	U8		intsToSet = 0;
	DWORD	dwReturn = ApiSuccess;

	#ifdef _DEBUG_INFO
	DebugMsg("AddCallbackCounts() started.");
	#endif

	DWORD dwWaitReason = WaitForSingleObject(g_srDllShared.hCBCountMutex, 1000L);
	if (dwWaitReason == WAIT_OBJECT_0) {

		#ifdef _DEBUG_INFO
		DebugMsg("AddCallbackCounts() received mutex");
		#endif

		__try {

			switch (cbToAdd) {

			case ASYNC_PERSEC:
				if (g_srDllShared.dllCallbackPerSecCnt == 0) {
					dwReturn = EnableTCRInterrupt(TCR_WRITE_PER_SECOND_INT_ENABLE);

					if (dwReturn == ApiSuccess)
					{
						#ifdef _DEBUG_INFO
						DebugMsg("AddCallbackCounts() ASYNC_PERSEC successfully added");
						#endif
					} else
					{
						#ifdef _DEBUG_ERROR
						DebugMsg("!!!AddCallbackCounts() ASYNC_PERSEC failed: %d", (int)dwReturn);
						#endif
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!AddCallbackCounts() ASYNC_PERSEC g_srDllShared.dllCallbackPerSecCnt not zero, ignore add");
					#endif
				}
				g_srDllShared.dllCallbackPerSecCnt++;
				break;

			case ASYNC_STARTTC:
				if (g_srDllShared.dllCallbackStartFrameCnt == 0) {
					dwReturn = EnableTCRInterrupt(TCR_WRITE_FRM_ST_INT_ENABLE);

					if (dwReturn == ApiSuccess)
					{
						#ifdef _DEBUG_INFO
						DebugMsg("AddCallbackCounts() ASYNC_STARTTC successfully added");
						#endif
					} else
					{
						#ifdef _DEBUG_ERROR
						DebugMsg("!!!AddCallbackCounts() ASYNC_STARTTC failed: %d", (int)dwReturn);
						#endif
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!AddCallbackCounts() ASYNC_STARTTC g_srDllShared.dllCallbackPerSecCnt not zero, ignore add");
					#endif
				}

				g_srDllShared.dllCallbackStartFrameCnt++;
				break;

			case ASYNC_ENDTC:
				if (g_srDllShared.dllCallbackEndFrameCnt == 0) {
					dwReturn = EnableTCRInterrupt(TCR_WRITE_FRM_END_INT_ENABLE);

					if (dwReturn == ApiSuccess)
					{
						#ifdef _DEBUG_INFO
						DebugMsg("AddCallbackCounts() ASYNC_ENDTC successfully added");
						#endif
					} else
					{
						#ifdef _DEBUG_ERROR
						DebugMsg("!!!AddCallbackCounts() ASYNC_ENDTC failed: %d", (int)dwReturn);
						#endif
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!AddCallbackCounts() ASYNC_ENDTC g_srDllShared.dllCallbackPerSecCnt not zero, ignore add");
					#endif
				}

				g_srDllShared.dllCallbackEndFrameCnt++;
				break;

			case ASYNC_TIMESET:
				if (g_srDllShared.dllCallbackPerMinCnt == 0) {
					dwReturn = EnableTCRInterrupt(TCR_WRITE_PER_MINUTE_INT_ENABLE);

					if (dwReturn == ApiSuccess)
					{
						#ifdef _DEBUG_INFO
						DebugMsg("AddCallbackCounts() ASYNC_TIMESET successfully added");
						#endif
					} else
					{
						#ifdef _DEBUG_ERROR
						DebugMsg("!!!AddCallbackCounts() ASYNC_TIMESET failed: %d", (int)dwReturn);
						#endif
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!AddCallbackCounts() ASYNC_TIMESET g_srDllShared.dllCallbackPerSecCnt not zero, ignore add");
					#endif
				}

				g_srDllShared.dllCallbackPerMinCnt++;
				break;

			case ASYNC_TIMEGAIN:
				if (g_srDllShared.dllCallbackTCDetectCnt == 0) {
					dwReturn = EnableTCRInterrupt(TCR_WRITE_TC_GAIN_INT_ENABLE);

					if (dwReturn == ApiSuccess)
					{
						#ifdef _DEBUG_INFO
						DebugMsg("AddCallbackCounts() ASYNC_TIMEGAIN successfully added");
						#endif
					} else
					{
						#ifdef _DEBUG_ERROR
						DebugMsg("!!!AddCallbackCounts() ASYNC_TIMEGAIN failed: %d", (int)dwReturn);
						#endif
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!AddCallbackCounts() ASYNC_TIMEGAIN g_srDllShared.dllCallbackPerSecCnt not zero, ignore add");
					#endif
				}

				g_srDllShared.dllCallbackTCDetectCnt++;
				break;

			case ASYNC_TIMELOSS:
				if (g_srDllShared.dllCallbackTCLossCnt == 0) {
					dwReturn = EnableTCRInterrupt(TCR_WRITE_TC_LOSS_INT_ENABLE);

					if (dwReturn == ApiSuccess)
					{
						#ifdef _DEBUG_INFO
						DebugMsg("AddCallbackCounts() ASYNC_TIMELOSS successfully added");
						#endif
					} else
					{
						#ifdef _DEBUG_ERROR
						DebugMsg("!!!AddCallbackCounts() ASYNC_TIMELOSS failed: %d", (int)dwReturn);
						#endif
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!AddCallbackCounts() ASYNC_TIMELOSS g_srDllShared.dllCallbackPerSecCnt not zero, ignore add");
					#endif
				}

				g_srDllShared.dllCallbackTCLossCnt++;
				break;

			default:
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!AddCallbackCounts() unknown callback: %d", (int)cbToAdd);
				#endif
				break;
			}
		}
		__finally {
			#ifdef _DEBUG_INFO1
						DebugMsg("Add %d %d %d %d %d %d",
									g_srDllShared.dllCallbackPerSecCnt, g_srDllShared.dllCallbackStartFrameCnt, g_srDllShared.dllCallbackEndFrameCnt,
									g_srDllShared.dllCallbackPerMinCnt, g_srDllShared.dllCallbackTCDetectCnt, g_srDllShared.dllCallbackTCLossCnt);
			#endif
			ReleaseMutex(g_srDllShared.hCBCountMutex);
		}
	}
	else
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!AddCallbackCounts() mutex not received: %ld", (unsigned long)dwWaitReason);
		#endif
	}


	if (dwReturn == ApiSuccess)
	{
		return ASYNC_OKAY;
	}
	else
	{
		return ASYNC_MASK_ERROR;
	}
}

//
//		Name:			RemoveCallbackCounts
//
//		Description:	Adjusts the callback count based on removing the callbacks
//						specified in the cbToRemove parameter.
//
//		Returns the interrupts left, or ASYNC_MASK_ERROR if there is an error.
//

DWORD RemoveCallbackCounts(DWORD cbToRemove)
{
	U8		intsToSet = 0;
	DWORD	dwReturn = ApiSuccess;

	#ifdef _DEBUG_INFO
	DebugMsg("RemoveCallbackCounts() started");
	#endif

	DWORD dwWaitReason = WaitForSingleObject(g_srDllShared.hCBCountMutex, 1000L);
	if (dwWaitReason == WAIT_OBJECT_0) {

		__try {

			if (cbToRemove & ASYNC_PERSEC) {
				if (g_srDllShared.dllCallbackPerSecCnt)
				{
					g_srDllShared.dllCallbackPerSecCnt--;
					if (g_srDllShared.dllCallbackPerSecCnt == 0) {
						dwReturn = DisableTCRInterrupt(TCR_WRITE_PER_SECOND_INT_ENABLE);
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!Remove callback ASYNC_PERSEC specified, but no callback found");
					#endif
				}
			}

			if (cbToRemove & ASYNC_STARTTC) {
				if (g_srDllShared.dllCallbackStartFrameCnt)
				{
					g_srDllShared.dllCallbackStartFrameCnt--;
					if (g_srDllShared.dllCallbackStartFrameCnt == 0) {
						dwReturn = DisableTCRInterrupt(TCR_WRITE_FRM_ST_INT_ENABLE);
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!Remove callback ASYNC_STARTTC specified, but no callback found");
					#endif
				}
			}

			if (cbToRemove & ASYNC_ENDTC) {
				if (g_srDllShared.dllCallbackEndFrameCnt)
				{
					g_srDllShared.dllCallbackEndFrameCnt--;
					if (g_srDllShared.dllCallbackEndFrameCnt == 0) {
						dwReturn = DisableTCRInterrupt(TCR_WRITE_FRM_END_INT_ENABLE);
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!Remove callback ASYNC_ENDTC specified, but no callback found");
					#endif
				}
			}

			if (cbToRemove & ASYNC_TIMESET) {
				if (g_srDllShared.dllCallbackPerMinCnt)
				{
					g_srDllShared.dllCallbackPerMinCnt--;
					if (g_srDllShared.dllCallbackPerMinCnt == 0) {
						dwReturn = DisableTCRInterrupt(TCR_WRITE_PER_MINUTE_INT_ENABLE);
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!Remove callback ASYNC_TIMESET specified, but no callback found");
					#endif
				}
			}

			if (cbToRemove & ASYNC_TIMEGAIN) {
				if (g_srDllShared.dllCallbackTCDetectCnt)
				{
					g_srDllShared.dllCallbackTCDetectCnt--;
					if (g_srDllShared.dllCallbackTCDetectCnt == 0) {
						dwReturn = DisableTCRInterrupt(TCR_WRITE_TC_GAIN_INT_ENABLE);
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!Remove callback ASYNC_TIMEGAIN specified, but no callback found");
					#endif
				}
			}

			if (cbToRemove & ASYNC_TIMELOSS) {
				if (g_srDllShared.dllCallbackTCLossCnt)
				{
					g_srDllShared.dllCallbackTCLossCnt--;
					if (g_srDllShared.dllCallbackTCLossCnt == 0) {
						dwReturn = DisableTCRInterrupt(TCR_WRITE_TC_LOSS_INT_ENABLE);
					}
				} else
				{
					#ifdef _DEBUG_ERROR
					DebugMsg("!!!Remove callback ASYNC_TIMELOSS specified, but no callback found");
					#endif
				}
			}
		}

		__finally {
			//#ifdef _DEBUG_ERROR
			//			DebugMsg("Rem %d %d %d %d %d %d",
			//						(int)g_srDllShared.dllCallbackPerSecCnt, (int)g_srDllShared.dllCallbackStartFrameCnt, (int)g_srDllShared.dllCallbackEndFrameCnt,
			//						(int)g_srDllShared.dllCallbackPerMinCnt, (int)g_srDllShared.dllCallbackTCDetectCnt, (int)g_srDllShared.dllCallbackTCLossCnt);
			//#endif
			ReleaseMutex(g_srDllShared.hCBCountMutex);
		}
	}
	else
	{
		dwReturn = ~ApiSuccess;
	}

	if (dwReturn == ApiSuccess)
	{
		return ASYNC_OKAY;
	}
	else
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!RemoveCallbackCounts() failed:%d", (int)dwReturn);
		#endif
		return ASYNC_MASK_ERROR;
	}

}


//
//		Name:			WriteInterruptStatus
//
//		Description:	Writes the interupts that need to be enabled to the card.
//
//		Returns 0 for success, 1 for fail.
//

DWORD WriteInterruptStatus(U8 intsToSet)
{
	PLX_STATUS status;

	//status = TCRWrite(0x00, &intsToSet, 1, BitSize8);
	status = PlxPci_PciBarSpaceWriteDynamic(&hPlxDevice, 2, 0x00, &intsToSet, 1, BitSize8, FALSE);

	if (status == ApiSuccess)
		return 0;
	else
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!WriteInterruptStatus: failed: %d",(int)status);
		#endif
		return 1;
	}
}

//
//		Name:			GetInterruptData
//
//		Description:	If this is a new interrupt, get the timeframe from the card.
//						Otherwise, return the stored timeframe.
//
//		Returns 0 for success, 1 for fail.
//

DWORD GetInterruptData(PUCHAR pIntReason, PTIMEFRAME pTimeFrame)
{
	DWORD dwReturn;

	if (!pIntReason || !pTimeFrame)
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!GetInterruptData: failed");
		#endif
		return 1;	// data error
	}
	
	// Acquire the mutex so we can access the LastInt data.
	// Then get the count. If the count is different from
	// the stored count, then get the interrupt status; otherwise,
	// use the shared interrupt status.

	// If a new interrupt and the reason indicates a timing interrupt
	// also get the data and validate it.

	DWORD dwWait = WaitForSingleObject(g_srDllShared.hLastIntMutex, 1000L);
	DWORD dwrc = 2;
	if (dwWait != WAIT_OBJECT_0)
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!GetInterruptData: failed to acquire mutex: %d",(int)dwWait);
		#endif

		return 1;	// Failed to acquire mutex
	}

	// __try ... __finally is used so the mutex is released no matter what might happen
	// (reference: "Using Mutex Objects" in platform SDK docs)

	__try {
		LARGE_INTEGER liPerfCount;
		RETURN_CODE rc;

	// kjf xxx yyy zzz why are these here? remove them
	//unsigned char VersionMajor, VersionMinor, VersionRevision; 
	//PlxPci_ApiVersion(&VersionMajor,&VersionMinor,&VersionRevision);

		// This is a call into the PLX DLL that we added
		rc = McIntrCountGetDynamic(&hPlxDevice, &liPerfCount);

		if (liPerfCount.QuadPart != g_srDllShared.dllLastIntPerfCount.QuadPart)
		{
			//#ifdef _DEBUG_INFO
			//DebugMsg("GetInterruptData: retry due to quad change");
			//#endif

			// There's been a new interrupt. Save the count, and get
			// the new status and data from the card.

			g_srDllShared.dllLastIntReason = 0;
			//rc = TCRRead(TCR_READ_INT_REASON, &g_srDllShared.dllLastIntReason, 1, BitSize8);
			rc = PlxPci_PciBarSpaceReadDynamic(&hPlxDevice, 2, TCR_READ_INT_REASON, &g_srDllShared.dllLastIntReason, 1, BitSize8, FALSE);

			if (rc == ApiSuccess)
			{

				g_srDllShared.dllLastIntPerfCount.QuadPart = liPerfCount.QuadPart;

				if (g_srDllShared.dllLastIntReason & ASYNC_TIMING_INTS) {
					dwReturn = PopulateTimeFrame(&g_srDllShared.dllLastIntTimeframe, liPerfCount);
					if (dwReturn)
					{
						#ifdef _DEBUG_ERROR
						DebugMsg("!!!GetInterruptData: failed populate: %d",(int)dwReturn);
						#endif
						dwrc = 1;	// failed reading data
					}
					else
						dwrc = 0;
				}
				else
					dwrc = 0;
			}
			else
			{
				#ifdef _DEBUG_ERROR
				DebugMsg("!!!GetInterruptData: failed read: %d",(int)rc);
				#endif
				dwrc = 1;
			}
		}
		else
		{
			dwrc = 0;
		}

		*pIntReason = g_srDllShared.dllLastIntReason;
		CopyMemory(pTimeFrame, &g_srDllShared.dllLastIntTimeframe, sizeof(TIMEFRAME)); 
	}
	__finally {
		ReleaseMutex(g_srDllShared.hLastIntMutex);
	}

	return dwrc;
}

//
//		Name:			EnableTCRInterrupt
//
//		Description:	Enable the passed in TCR interrupt
//
//		Returns 0 for success, 1 for fail.
//

RETURN_CODE EnableTCRInterrupt(DWORD intToEnable)
{
	RETURN_CODE rc;
	U8	val = 0xFF;

	#ifdef _DEBUG_INFO
	DebugMsg("Enabling int #%d", intToEnable);
	#endif

	//rc = TCRWrite(intToEnable, &val, 1, BitSize8);
	rc = PlxPci_PciBarSpaceWriteDynamic(&hPlxDevice, 2, intToEnable, &val, 1, BitSize8, FALSE);

	if ( rc == ApiSuccess)
	{
		#ifdef _DEBUG_INFO
		DebugMsg("EnableTCRInterrupt: successful");
		#endif
	} else
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!EnableTCRInterrupt: failed enable: %d", (int)rc);
		#endif
	}

	return rc;
}

//
//		Name:			DisableTCRInterrupt
//
//		Description:	Disable the passed in TCR interrupt
//
//		Returns 0 for success, 1 for fail.
//

RETURN_CODE DisableTCRInterrupt(DWORD intToDisable)
{
	RETURN_CODE rc;
	U8	val = 0;

	#ifdef _DEBUG_INFO
	DebugMsg("Disabling int #%d", intToDisable);
	#endif

	//rc = TCRWrite(intToEnable, &val, 1, BitSize8);
	rc = PlxPci_PciBarSpaceWriteDynamic(&hPlxDevice, 2, intToDisable, &val, 1, BitSize8, FALSE);

	if ( rc == ApiSuccess)
	{
		#ifdef _DEBUG_INFO
		DebugMsg("DisableTCRInterrupt: successful");
		#endif
	} else
	{
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!DisableTCRInterrupt: failed disable: %d", (int)rc);
		#endif
	}

	return rc;
}

	






//
// This function load the PLX DLL into memory
// When using a CLR based project, the DLL must be loaded dynamically.
//
boolean fnLoadPlxDll(HINSTANCE *hinstLib)  // must use double indirection
{
	HINSTANCE hLocalInstLib;

	boolean bLoadIsSuccessful = true;

	hLocalInstLib = NULL;

	// kjf xxx yyy zzz this will need to use a path determined at run time or, 
	// we must force the end user to install the DLL into a known location


	#ifdef _DEBUG
		#ifdef _Masterclock_x64 // defined in the project settings, C++ preprocessor
			//CFileFind cffFindDriver;
			//BOOL bFound = cffFindDriver.FindFile("C:\\Plx\\PlxSdk\\Windows\\PlxApi\\Output\\x64\\Debug\\PlxApi640_x64_rev_a.dll");
			//if (bFound)
			//{

			WIN32_FIND_DATA FindFileData;
			HANDLE hFind;

			hFind = FindFirstFile("C:\\Plx\\PlxSdk\\Windows\\PlxApi\\Output\\x64\\Debug\\PlxApi640_x64_rev_a.dll", &FindFileData);
			if (hFind == INVALID_HANDLE_VALUE) 
			{
				DebugMsg("!!!FindFirstFile failed (%d)\n", GetLastError());
				return false;
			} 
			else 
			{
				DebugMsg("The first file found is %s\n", FindFileData.cFileName);

				FindClose(hFind);
			}

			hLocalInstLib = LoadLibrary("C:\\Plx\\PlxSdk\\Windows\\PlxApi\\Output\\x64\\Debug\\PlxApi640_x64_rev_a.dll"); 
		#else
			hLocalInstLib = LoadLibrary("C:\\Plx\\PlxSdk\\Windows\\PlxApi\\Output\\Win32\\Debug\\PlxApi640_rev_a.dll");  
		#endif
	#else
		#ifdef _Masterclock_x64 // defined in the project settings, C++ preprocessor
			hLocalInstLib = LoadLibrary(".\\drivers\\PlxApi640_x64_rev_a.dll"); 
		#else 
			hLocalInstLib = LoadLibrary(".\\drivers\\PlxApi640_rev_a.dll"); 
		#endif
	#endif
	
	// This is done using hLocalInstLib and the line below to make it easier to understand the double indirection
	*hinstLib     = hLocalInstLib;
	 
	// If the handle is valid, try to get the function address.
	 
	if (*hinstLib == NULL) 
	{ 
		// error loading the DLL
		bLoadIsSuccessful = false;
	} else
	{

		#ifdef _DEBUG_ERROR
		DebugMsg("TCR API delay 2 sec to allow PLX.DLL to complete initialization");
		#endif
		Sleep(2000);
		#ifdef _DEBUG_ERROR
		DebugMsg("TCR API dll 2 sec delay completed");
		#endif
	}

	return bLoadIsSuccessful;
}

//
// This function unloads the PLX DLL from memory
// When using a CLR based project, the DLL must be loaded dynamically.
//
boolean fnFreePlxDll(HINSTANCE hinstLib)
{
	BOOL fFreeIsSuccessful=true;

	// Free the DLL module.

	fFreeIsSuccessful = FreeLibrary(hinstLib);

	if (!fFreeIsSuccessful)
	{
		// error freeing DLL
	}

	return fFreeIsSuccessful;
}






// All of the calls into the PLX DLL must be done through the code below.
// The PLX DLL cannot be a statically linked library. It must be dynamically linked.
//PLX_STATUS PlxPci_DeviceOpen(PLX_DEVICE_KEY *pKey, PLX_DEVICE_OBJECT *pDevice);
typedef PLX_STATUS (*MYPROC1)(PLX_DEVICE_KEY *pKey, PLX_DEVICE_OBJECT *pDevice);

PLX_STATUS PlxPci_DeviceOpenDynamic(PLX_DEVICE_KEY *pKey, PLX_DEVICE_OBJECT *pDevice)
{
	PLX_STATUS plxStatus;
	
	MYPROC1 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 
	
	// Get a handle to the DLL module.
	
	plxStatus = ApiFailed;
	
	// If the handle is valid, try to get the function address.
	
	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
	    ProcAdd = (MYPROC1) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_DeviceOpen"); 
	
		// If the function address is valid, call the function.
	
	    if (ProcAdd != NULL) 
	    {
	        fRunTimeLinkSuccess = TRUE;
	        plxStatus = (ProcAdd)(pKey, pDevice); // this is the dll call 
	    }
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}





typedef PLX_STATUS (*MYPROC2)(PLX_DEVICE_OBJECT *pDevice);

PLX_STATUS PlxPci_DeviceCloseDynamic(PLX_DEVICE_OBJECT *pDevice)
{
	PLX_STATUS plxStatus;

	MYPROC2 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC2) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_DeviceClose"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef PLX_STATUS (*MYPROC3)(PLX_DEVICE_KEY *pKey, U8 DeviceNumber);

PLX_STATUS PlxPci_DeviceFindDynamic(PLX_DEVICE_KEY *pKey, U8 DeviceNumber)
{
	PLX_STATUS plxStatus;

	MYPROC3 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC3) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_DeviceFind"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pKey, DeviceNumber); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef PLX_STATUS (*MYPROC4)(PLX_DEVICE_OBJECT *pDevice, PLX_DRIVER_PROP *pDriverProp);

PLX_STATUS PlxPci_DriverPropertiesDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_DRIVER_PROP *pDriverProp)
{
	PLX_STATUS plxStatus;

	MYPROC4 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC4) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_DriverProperties"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, pDriverProp); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef PLX_STATUS (*MYPROC5)(PLX_DEVICE_OBJECT *pDevice, U8 *pVersionMajor, U8 *pVersionMinor, U8 *pVersionRevision);

PLX_STATUS PlxPci_DriverVersionDynamic(PLX_DEVICE_OBJECT *pDevice, U8 *pVersionMajor, U8 *pVersionMinor, U8 *pVersionRevision)
{
	PLX_STATUS plxStatus;

	MYPROC5 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC5) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_DriverVersion"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, pVersionMajor, pVersionMinor, pVersionRevision); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef U32 (*MYPROC6)(PLX_DEVICE_OBJECT *pDevice, U16 offset, PLX_STATUS *pStatus);

U32 PlxPci_PciRegisterReadFastDynamic(PLX_DEVICE_OBJECT *pDevice, U16 offset, PLX_STATUS *pStatus)
{
	U32 plxStatus;

	MYPROC6 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC6) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_PciRegisterReadFast"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, offset, pStatus); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef PLX_STATUS (*MYPROC7)(PLX_DEVICE_OBJECT *pDevice, U8 BarIndex, U32 offset, VOID *pBuffer, U32 ByteCount, PLX_ACCESS_TYPE AccessType, BOOLEAN bOffsetAsLocalAddr);

PLX_STATUS PlxPci_PciBarSpaceReadDynamic(PLX_DEVICE_OBJECT *pDevice, U8 BarIndex, U32 offset, VOID *pBuffer, U32 ByteCount, PLX_ACCESS_TYPE AccessType, BOOLEAN bOffsetAsLocalAddr)
{
	PLX_STATUS plxStatus;

	MYPROC7 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC7) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_PciBarSpaceRead"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, BarIndex, offset, pBuffer, ByteCount, AccessType, bOffsetAsLocalAddr); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef PLX_STATUS (*MYPROC8)(PLX_DEVICE_OBJECT *pDevice, U8 BarIndex, U32 offset, VOID *pBuffer, U32 ByteCount, PLX_ACCESS_TYPE AccessType, BOOLEAN bOffsetAsLocalAddr);

PLX_STATUS PlxPci_PciBarSpaceWriteDynamic(PLX_DEVICE_OBJECT *pDevice, U8 BarIndex, U32 offset, VOID *pBuffer, U32 ByteCount, PLX_ACCESS_TYPE AccessType, BOOLEAN bOffsetAsLocalAddr)
{
	PLX_STATUS plxStatus;

	MYPROC8 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC8) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_PciBarSpaceWrite"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, BarIndex, offset, pBuffer, ByteCount, AccessType, bOffsetAsLocalAddr); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef PLX_STATUS (*MYPROC9)(PLX_DEVICE_OBJECT *pDevice, PLX_INTERRUPT *pPlxIntr);

PLX_STATUS PlxPci_InterruptEnableDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_INTERRUPT *pPlxIntr)
{
	PLX_STATUS plxStatus;

	MYPROC9 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC9) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_InterruptEnable"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, pPlxIntr); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef PLX_STATUS (*MYPROC10)(PLX_DEVICE_OBJECT *pDevice, PLX_INTERRUPT *pPlxIntr);

PLX_STATUS PlxPci_InterruptDisableDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_INTERRUPT *pPlxIntr)
{
	PLX_STATUS plxStatus;

	MYPROC10 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC10) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_InterruptDisable"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, pPlxIntr); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef PLX_STATUS (*MYPROC11)(PLX_DEVICE_OBJECT *pDevice, PLX_INTERRUPT *pPlxIntr, PLX_NOTIFY_OBJECT *pEvent);

PLX_STATUS PlxPci_NotificationRegisterForDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_INTERRUPT *pPlxIntr, PLX_NOTIFY_OBJECT *pEvent)
{
	PLX_STATUS plxStatus;

	MYPROC11 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC11) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_NotificationRegisterFor"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, pPlxIntr, pEvent); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef PLX_STATUS (*MYPROC12)(PLX_DEVICE_OBJECT *pDevice, PLX_NOTIFY_OBJECT *pEvent, U64 Timeout_ms);

PLX_STATUS PlxPci_NotificationWaitDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_NOTIFY_OBJECT *pEvent, U64 Timeout_ms)
{
	PLX_STATUS plxStatus;

	MYPROC12 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC12) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_NotificationWait"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, pEvent, Timeout_ms); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef PLX_STATUS (*MYPROC13)(PLX_DEVICE_OBJECT *pDevice, PLX_NOTIFY_OBJECT *pEvent, PLX_INTERRUPT *pPlxIntr);

PLX_STATUS PlxPci_NotificationStatusDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_NOTIFY_OBJECT *pEvent, PLX_INTERRUPT *pPlxIntr)
{
	PLX_STATUS plxStatus;

	MYPROC13 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC13) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_NotificationStatus"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, pEvent, pPlxIntr); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef PLX_STATUS (*MYPROC14)(PLX_DEVICE_OBJECT *pDevice, PLX_NOTIFY_OBJECT *pEvent);

PLX_STATUS PlxPci_NotificationCancelDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_NOTIFY_OBJECT *pEvent)
{
	PLX_STATUS plxStatus;

	MYPROC14 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC14) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_NotificationCancel"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, pEvent); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef PLX_STATUS (*MYPROC15)(PLX_DEVICE_OBJECT *pDevice, PLX_PORT_PROP *pPortProp);

PLX_STATUS PlxPci_GetPortPropertiesDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_PORT_PROP *pPortProp)
{
	PLX_STATUS plxStatus;

	MYPROC15 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC15) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "PlxPci_GetPortProperties"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, pPortProp); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}







typedef RETURN_CODE (*MYPROC16)(PLX_DEVICE_OBJECT *pDevice, LARGE_INTEGER *pliPerformanceCountAtIntr);

RETURN_CODE McIntrCountGetDynamic(PLX_DEVICE_OBJECT *pDevice, LARGE_INTEGER *pliPerformanceCountAtIntr)
{
	RETURN_CODE plxStatus;

	MYPROC16 ProcAdd; 
	BOOL fRunTimeLinkSuccess = FALSE; 

	// Get a handle to the DLL module.

	plxStatus = ApiFailed;

	// If the handle is valid, try to get the function address.

	if (g_srDllShared.hPlxDllInstanceLib != NULL) 
	{ 
		ProcAdd = (MYPROC16) GetProcAddress(g_srDllShared.hPlxDllInstanceLib, "McIntrCountGet"); 

		// If the function address is valid, call the function.

		if (ProcAdd != NULL) 
		{
			fRunTimeLinkSuccess = TRUE;
			plxStatus = (ProcAdd)(pDevice, pliPerformanceCountAtIntr); // this is the dll call 
		}
	} 

	// If unable to call the DLL function, use an alternative.
	//if (! fRunTimeLinkSuccess) 
	//    printf("Message printed from executable\n"); 

	return plxStatus;
}






