#pragma once

#include "SoftwareConfig.h"
#include "drivercommon.h"
#include "tcrcommon.h"
#include "IntApi.h"


void AcquireApiMutex(void);
void ReleaseApiMutex(void);

// DLL definitions

#define DLL_VERSION_MAJ			2
#define DLL_VERSION_MIN			1
#define DLL_VERSION_REV			0

// Macros

#define TCR_WRITE_PER_MINUTE_INT_ENABLE	0x00
#define TCR_WRITE_TC_LOSS_INT_ENABLE	0x01
#define TCR_WRITE_TC_GAIN_INT_ENABLE	0x02
#define TCR_WRITE_FRM_END_INT_ENABLE	0x03
#define TCR_WRITE_FRM_ST_INT_ENABLE		0x04
#define TCR_WRITE_PER_SECOND_INT_ENABLE	0x05
#define TCR_WRITE_RECALIBRATE			0x06

#define TCR_READ_DATA_BUFFER		0x00
#define TCR_READ_DATA_CHKSUM		0x08
#define TCR_READ_TC_STATUS			0x09
#define TCR_READ_DIG_POT_SETTING	0x0A
#define TCR_READ_INT_REASON			0x0B


//#define TCRWrite(loc, ptr, numBytes, size) \
//	PlxPci_PciBarSpaceWriteDynamic(&hPlxDevice, 0, (loc), (ptr), (numBytes), (size), FALSE)
//#define TCRRead(loc, ptr, numBytes, size) \
//	PlxPci_PciBarSpaceReadDynamic(&hPlxDevice, 0, (loc), (ptr), (numBytes), (size), FALSE)

//
// a structure to contain a time code frame read and
// associated statuses
//

typedef struct _TIMEFRAME
{
	UCHAR			tcBuffer[8];	// Buffer for timecode to be read into
	UCHAR			tcChecksum;		// timecode checksum
	UCHAR			tcStatus;		// time/time code status

    LARGE_INTEGER   intPerfCount;   // performance counter as of last interrupt

    DWORD           hwStatus;				// hardware status

} TIMEFRAME, *PTIMEFRAME;



// time code reader hardware and kernel-mode driver status information

typedef struct _DEVICE_INFO
{
	// model ID of TCR card in use, one of TC_PRODUCT_xxx
	DWORD		devType;

	// model version of TCR card in use (not applicable for TCR-200/500 so currently unused)
	DWORD		devVersion;

	// operation mode of TCR card (for TCR-500, always read, for TCR-200 read, generate, or stopped)
	DWORD		devMode; // not implemented

	// status of TCR card, one of HW_STATUS_xxx from TCRPUB.H
	DWORD		devStatus;

	// TCR card kernel-mode driver version
	TCRDRVVER	driverVersion;
	
	// TCR card kernel-mode driver status, one of TC_STATUS_xxx from TCRPUB.H
	DWORD		tcStatus;

	// GPS status (if time code source has SMPTE, Leitch encoding, with Masterclock, Inc proprietary control bits)
	DWORD		gpsStatus;

	// frame rate of incoming time code (only used for time codes with sub-second frame rates)
	WORD		frameRate;

} DEVICE_INFO, *PDEVICE_INFO;

typedef struct _CALLBACK_NODE
{
	// callback control
	PTCRASYNC_CALLBACK_ROUTINE	cbFunction;
	DWORD						cbThreadId;
	HANDLE						cbThreadHandle;
	DWORD						cbHooksMask;
	PVOID						next;
	PVOID						prev;
	BOOL						locked;

	// data for each thread's callback function
	TCRTIME						timeData;
} CALLBACK_NODE, *PCALLBACK_NODE;

extern BOOL ProcessTimeFrame(TIMEFRAME *t, TCRTIME *tcrt);
extern DWORD PopulateTimeFrame(PTIMEFRAME pTimeFrame, LARGE_INTEGER liPerfCount);
extern DWORD GetTCTypeFromTCStatus(U8 tcStatus);

extern DWORD			dwTcType;
extern DWORD			dllInitStatus;
extern DEVICE_INFO		devInfo;
extern PCALLBACK_NODE	cbListHead;
//extern HANDLE			tcrInterruptEvent;

// interprocess DLL instance shared data and handling functions
typedef struct {
	HINSTANCE hPlxDllInstanceLib; 
	volatile LONG  ldllApiMutex;		// We cannot use a true mutex (ie. CreateMutex() and WaitForSingleObject() )
	// Simulate one using this flag.

	DWORD		dllInstanceCnt;

	// Critical section for accessing Callback list data
	CRITICAL_SECTION csCBListData;

	// Mutex to control access to Callback counts
	HANDLE	    hCBCountMutex;
	DWORD		dllCallbackPerSecCnt;
	DWORD		dllCallbackStartFrameCnt;
	DWORD		dllCallbackEndFrameCnt;
	DWORD		dllCallbackPerMinCnt;
	DWORD		dllCallbackTCDetectCnt;
	DWORD		dllCallbackTCLossCnt;

	// Mutex to control access to LastInt status and perf count.
	HANDLE	      hLastIntMutex;
	UCHAR		  dllLastIntReason;
	TIMEFRAME	  dllLastIntTimeframe; 
	LARGE_INTEGER dllLastIntPerfCount;
} MC_DLL_SHARED;

extern MC_DLL_SHARED g_srDllShared;

extern DWORD RemoveCallbackCounts(DWORD cbToRemove);
extern DWORD AddCallbackCounts(DWORD cbToAdd);


PLX_STATUS PlxPci_DeviceOpenDynamic(PLX_DEVICE_KEY *pKey, PLX_DEVICE_OBJECT *pDevice);
PLX_STATUS PlxPci_DeviceCloseDynamic(PLX_DEVICE_OBJECT *pDevice);
PLX_STATUS PlxPci_DeviceFindDynamic(PLX_DEVICE_KEY *pKey, U8 DeviceNumber);
PLX_STATUS PlxPci_DriverPropertiesDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_DRIVER_PROP *pDriverProp);
PLX_STATUS PlxPci_DriverVersionDynamic(PLX_DEVICE_OBJECT *pDevice, U8 *pVersionMajor, U8 *pVersionMinor, U8 *pVersionRevision);
U32 PlxPci_PciRegisterReadFastDynamic(PLX_DEVICE_OBJECT *pDevice, U16 offset, PLX_STATUS *pStatus);
PLX_STATUS PlxPci_PciBarSpaceReadDynamic(PLX_DEVICE_OBJECT *pDevice, U8 BarIndex, U32 offset, VOID *pBuffer, U32 ByteCount, PLX_ACCESS_TYPE AccessType, BOOLEAN bOffsetAsLocalAddr);
PLX_STATUS PlxPci_PciBarSpaceWriteDynamic(PLX_DEVICE_OBJECT *pDevice, U8 BarIndex, U32 offset, VOID *pBuffer, U32 ByteCount, PLX_ACCESS_TYPE AccessType, BOOLEAN bOffsetAsLocalAddr);
PLX_STATUS PlxPci_InterruptEnableDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_INTERRUPT *pPlxIntr);
PLX_STATUS PlxPci_InterruptDisableDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_INTERRUPT *pPlxIntr);
PLX_STATUS PlxPci_NotificationRegisterForDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_INTERRUPT *pPlxIntr, PLX_NOTIFY_OBJECT *pEvent);
PLX_STATUS PlxPci_NotificationWaitDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_NOTIFY_OBJECT *pEvent, U64 Timeout_ms);
PLX_STATUS PlxPci_NotificationStatusDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_NOTIFY_OBJECT *pEvent, PLX_INTERRUPT *pPlxIntr);
PLX_STATUS PlxPci_NotificationCancelDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_NOTIFY_OBJECT *pEvent);
PLX_STATUS PlxPci_GetPortPropertiesDynamic(PLX_DEVICE_OBJECT *pDevice, PLX_PORT_PROP *pPortProp);
RETURN_CODE McIntrCountGetDynamic(PLX_DEVICE_OBJECT *pDevice, LARGE_INTEGER *pliPerformanceCountAtIntr);
