#include "StdAfx.h"

#include <stdio.h>
#include <string.h>
#include "tcrcommon.h"
#include "SoftwareConfig.h"
#include "DebugOutput.h"

//#pragma managed

// registry key names
#define REGISTRY_TIME_ZONE_KEY         ("Registry Time Zone Key")
#define REGISTRY_TIME_ZONE_DESC        ("Source Time Zone Description")
#define REGISTRY_TIME_ZONE_INFO        ("Source Time Zone")
#define REGISTRY_DATE_AVAIL            ("Date Available")
#define REGISTRY_DST_ENCODED           ("DST Encoded")
#define REGISTRY_TIME_CODE_SOURCE      ("Time Code Source")
#define REGISTRY_FREEWHEEL_SYNC        ("Freewheel Synchronize")
#define REGISTRY_FREEWHEEL_DISP        ("Freewheel Display")
#define REGISTRY_TIME_CODE_LOSS        ("Time Code Loss Notify")
#define REGISTRY_AGGRESSIVE_TIME_SET   ("Aggressive Time Set")
#define REGISTRY_CRUDE_TIME_SET        ("Use Crude Time Set")

int tzCompare(const void *arg1, const void *arg2);


CSoftwareConfig::CSoftwareConfig()
{

	#ifdef _DEBUG_CONFIG
	DebugMsg("Load Config class");
	#endif

	GetKeyDefaults(&m_srKeys);

	// This will verify the keys exist in the registry. If they do not, the key will be created and set to the default value
	VerifyRegistryKeys();

	LoadConfigFromRegistry();

	GetRegistryTimeZones();
}


//
//		Name:				::GetImagePath()
//
//		Description:		Assumes the registry keys exist and have been initialized.
//

//void CSoftwareConfig::GetImagePath(char *destBuff, int iBuffSize)
//{
//	char	cParamPath[256];
//	HKEY	rKey;
//
//	DWORD	paramType;
//	DWORD	paramSize;
//
//	GetRegistryPath(cParamPath, sizeof(cParamPath));
//
//	RegOpenKeyEx(	HKEY_LOCAL_MACHINE,
//					cParamPath,
//					0,
//					KEY_READ,
//					&rKey);
//
//	paramType = REG_EXPAND_SZ;
//	paramSize = iBuffSize-1;
//
//	RegQueryValueEx(rKey,
//					"ImagePath",
//					NULL,
//					&paramType,
//					(unsigned char *)destBuff,
//					&paramSize);
//
//	RegCloseKey(rKey);
//}

//
//		Name:				::LoadConfigFromRegistry()
//
//		Description:		Loads software configuration settings from the Windows NT registry
//							from the \SYSTEM\CurrentControlSet\Services\ourservicename\Parameters
//							key.
//
//		Note:				Inability to load any of the undocumented settings does not fail the
//							configuration load.
//
//		Returns:			FALSE if error reading the registry.
//

int CSoftwareConfig::LoadConfigFromRegistry()
{
	HKEY	paramKey;
	DWORD	paramType;
	DWORD	paramSize;
	char	cParamPath[256];
	LSTATUS lQueryStatus;

	BOOL    bIsFailed;

	#ifdef _DEBUG_CONFIG
	DebugMsg("Load Config from registry");
	#endif
	bIsFailed = FALSE;

	GetRegistryPath(cParamPath, sizeof(cParamPath));

	strcat_s(cParamPath,sizeof(cParamPath), "\\Parameters");

	lQueryStatus = RegOpenKeyEx(HKEY_LOCAL_MACHINE,cParamPath,0,KEY_READ,&paramKey);
	if (lQueryStatus != ERROR_SUCCESS)
	{
		bIsFailed = TRUE; 
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Load Config path error:%d",(int)lQueryStatus);
		#endif
	}

	paramType = REG_SZ;
	paramSize = sizeof(m_srKeys.tzRegistryKey);

	lQueryStatus = RegQueryValueEx(paramKey,
		REGISTRY_TIME_ZONE_KEY,
		NULL,
		&paramType,
		(BYTE *)m_srKeys.tzRegistryKey,
		&paramSize);

	if (lQueryStatus != ERROR_SUCCESS)
	{
		bIsFailed = TRUE; 
        #ifdef _DEBUG_ERROR
		DebugMsg("!!!Load Config time zone key error:%d",(int)lQueryStatus);
        #endif
	}

	paramType = REG_SZ;
	paramSize = sizeof(m_srKeys.tzDesc);

	lQueryStatus = RegQueryValueEx(paramKey,
		REGISTRY_TIME_ZONE_DESC,
		NULL,
		&paramType,
		(BYTE *)m_srKeys.tzDesc,
		&paramSize);

	if (lQueryStatus != ERROR_SUCCESS)
	{
		bIsFailed = TRUE; 
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Load Config time zone description error:%d",(int)lQueryStatus);
		#endif
	}

	paramType = REG_BINARY;
	paramSize = sizeof(m_srKeys.tzInfo);

	lQueryStatus = RegQueryValueEx(paramKey,
		REGISTRY_TIME_ZONE_INFO,
		NULL,
		&paramType,
		(BYTE *)&m_srKeys.tzInfo,
		&paramSize);

	if (lQueryStatus != ERROR_SUCCESS)
	{
		bIsFailed = TRUE; 
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Load Config time zone info error:%d",(int)lQueryStatus);
		#endif
	}

	paramType = REG_DWORD;
	paramSize = 4;

	lQueryStatus = RegQueryValueEx(paramKey,
		REGISTRY_DATE_AVAIL,
		NULL,
		&paramType,
		(BYTE *)&m_srKeys.dateAvailable,
		&paramSize);

	if (lQueryStatus != ERROR_SUCCESS)
	{
		bIsFailed = TRUE; ;
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Load Config date avail error:%d",(int)lQueryStatus);
		#endif
	}

	paramType = REG_DWORD;
	paramSize = 4;
	
	lQueryStatus = RegQueryValueEx(paramKey,
		REGISTRY_DST_ENCODED,
		NULL,
		&paramType,
		(BYTE *)&m_srKeys.dstEncoded,
		&paramSize);

	if (lQueryStatus != ERROR_SUCCESS)
	{
		bIsFailed = TRUE; 
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Load Config dst encoded error error:%d",(int)lQueryStatus);
		#endif
	}

	paramType = REG_DWORD;
	paramSize = 4;
	
	lQueryStatus = RegQueryValueEx(paramKey,
		REGISTRY_TIME_CODE_SOURCE,
		NULL,
		&paramType,
		(BYTE *)&m_srKeys.timeCodeSource,
		&paramSize);

	if (lQueryStatus != ERROR_SUCCESS)
	{
		bIsFailed = TRUE; 
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Load Config time code source error:%d",(int)lQueryStatus);
		#endif
	}

	paramType = REG_DWORD;
	paramSize = 4;

	lQueryStatus = RegQueryValueEx(paramKey,
		REGISTRY_FREEWHEEL_SYNC,
		NULL,
		&paramType,
		(BYTE *)&m_srKeys.freewheelSynchronize,
		&paramSize);

	if (lQueryStatus != ERROR_SUCCESS)
	{
		bIsFailed = TRUE; 
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Load Config freewheel sync error:%d",(int)lQueryStatus);
		#endif
	}

	paramType = REG_DWORD;
	paramSize = 4;

	lQueryStatus = RegQueryValueEx(paramKey,
		REGISTRY_FREEWHEEL_DISP,
		NULL,
		&paramType,
		(BYTE *)&m_srKeys.freewheelDisplay,
		&paramSize);

	if (lQueryStatus != ERROR_SUCCESS)
	{
		bIsFailed = TRUE; 
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Load Config freewheel display error:%d",(int)lQueryStatus);
		#endif
	}

	paramType = REG_DWORD;
	paramSize = 4;

	lQueryStatus = RegQueryValueEx(paramKey,
		REGISTRY_TIME_CODE_LOSS,
		NULL,
		&paramType,
		(BYTE *)&m_srKeys.tcLossNotify,
		&paramSize);

	if (lQueryStatus != ERROR_SUCCESS)
	{
		bIsFailed = TRUE; 
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Load Config tc loss notify error:%d",(int)lQueryStatus);
		#endif
	}

	if (bIsFailed)
	{
		VerifyRegistryKeys();
	}

	// note that the remaining configuration items are undocumented or special-purpose
	// configuration and configuration load does not fail if they are missing

	//paramType = REG_DWORD;
	//paramSize = 4;

	//if (RegQueryValueEx(paramKey,
	//				REGISTRY_AGGRESSIVE_TIME_SET,
	//				NULL,
	//				&paramType,
	//				(BYTE *)&m_srKeys.setTimeAggressive,
	//				&paramSize) != ERROR_SUCCESS)
	//{
	//	m_srKeys.setTimeAggressive = 0L;
	//	#ifdef _DEBUG_ERROR
	//	DebugMsg("!!!Load Config aggressive set error");
	//	#endif
	//}

	//paramType = REG_DWORD;
	//paramSize = 4;

	//if (RegQueryValueEx(paramKey,
	//				REGISTRY_CRUDE_TIME_SET,
	//				NULL,
	//				&paramType,
	//				(BYTE *)&m_srKeys.useCrudeTimeSet,
	//				&paramSize) != ERROR_SUCCESS)
	//{
	//	m_srKeys.useCrudeTimeSet = 0L;
	//	#ifdef _DEBUG_ERROR
	//	DebugMsg("!!!Load Config crude set error");
	//	#endif
	//}

	RegCloseKey(paramKey);

	return(TRUE);
}


//
//		Name:				::SaveConfigToRegistry()
//
//		Description:		Saves software configuration members to the Windows NT registry
//							under the \SYSTEM\CurrentControlSet\Services\ourservicename\Parameters
//							key.
//
//		Note:				We don't write changes for the undocumented settings.  This is
//							a process to be performed only manually.
//
//		Returns:			FALSE if error writing to the registry.
//

BOOL CSoftwareConfig::SaveConfigToRegistry()
{
	HKEY	paramKey;
	char	cParamPath[256];

	#ifdef _DEBUG_CONFIG
	DebugMsg("Save Config to registry");
	#endif

	GetRegistryPath(cParamPath, sizeof(cParamPath));

	strcat_s(cParamPath,sizeof(cParamPath), "\\Parameters");

	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE,
		             cParamPath,
					 0,
					 KEY_SET_VALUE,
					 &paramKey) != ERROR_SUCCESS)
	{

		return FALSE;
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Save Config path error");
		#endif
	}

	RegSetValueEx( paramKey,
						REGISTRY_TIME_ZONE_KEY,
						0,
						REG_SZ,
						(BYTE *)m_srKeys.tzRegistryKey,
						(DWORD)strlen(m_srKeys.tzRegistryKey)+1);

	RegSetValueEx( paramKey,
						REGISTRY_TIME_ZONE_DESC,
						0,
						REG_SZ,
						(BYTE *)m_srKeys.tzDesc,
						(DWORD)strlen(m_srKeys.tzDesc)+1);

	RegSetValueEx(	paramKey,
						REGISTRY_TIME_ZONE_INFO,
						0,
						REG_BINARY,
						(BYTE *)&m_srKeys.tzInfo,
						(DWORD)sizeof(m_srKeys.tzInfo));

	RegSetValueEx( paramKey,
						REGISTRY_DATE_AVAIL,
						0,
						REG_DWORD,
						(BYTE *)&m_srKeys.dateAvailable,
						(DWORD)sizeof(DWORD));

	RegSetValueEx( paramKey,
						REGISTRY_DST_ENCODED,
						0,
						REG_DWORD,
						(BYTE *)&m_srKeys.dstEncoded,
						(DWORD)sizeof(DWORD));

	RegSetValueEx( paramKey,
						REGISTRY_TIME_CODE_SOURCE,
						0,
						REG_DWORD,
						(BYTE *)&m_srKeys.timeCodeSource,
						(DWORD)sizeof(DWORD));

	RegSetValueEx(	paramKey,
						REGISTRY_FREEWHEEL_SYNC,
						0,
						REG_DWORD,
						(BYTE *)&m_srKeys.freewheelSynchronize,
						(DWORD)sizeof(DWORD));

	RegSetValueEx(	paramKey,
						REGISTRY_FREEWHEEL_DISP,
						0,
						REG_DWORD,
						(BYTE *)&m_srKeys.freewheelDisplay,
						(DWORD)sizeof(DWORD));

	RegSetValueEx(	paramKey,
						REGISTRY_TIME_CODE_LOSS,
						0,
						REG_DWORD,
						(BYTE *)&m_srKeys.tcLossNotify,
						(DWORD)sizeof(DWORD));

	RegCloseKey(paramKey);

	return TRUE;
}



//
//		Name:				::GetRegistryPath()
//
//		Description:		Gets the registry path.
//

void CSoftwareConfig::GetRegistryPath(char *pPath, int iPathLength)
{
	sprintf_s(pPath,iPathLength,"SOFTWARE\\Masterclock\\%s",
		REGISTRY_SOURCE_NAME);
}

//
//		Name:			::GetKeyDefaults()
//
//		Description:	Set the structure to the default key values
//

void CSoftwareConfig::GetKeyDefaults(REGISTRY_KEYS *psrMyKeys)
{
	#ifdef _DEBUG_CONFIG
	DebugMsg("Get config key defaults");
	#endif

	memset(psrMyKeys, 0, sizeof(REGISTRY_KEYS));

	psrMyKeys->dstEncoded           = FALSE;
	psrMyKeys->timeCodeSource       = TC_SOURCE_DATE_INCLUDED;
	psrMyKeys->dateAvailable        = TRUE;
	psrMyKeys->freewheelSynchronize = TRUE;
	psrMyKeys->freewheelDisplay     = TRUE;
	psrMyKeys->tcLossNotify         = TRUE;
		
	// set up time zone defaults

	SetTzDescToUTC(psrMyKeys);

	// set defaults for undocumented settings

	//psrMyKeys->setTimeAggressive   = 0;
	//psrMyKeys->useCrudeTimeSet     = 0;
}

//
//		Name:			::VerifyRegistryKeys()
//
//		Description:	This will verify the keys exist in the registry.
//                      If the key does not exist, the key will be created and set to the default value.
//

void CSoftwareConfig::VerifyRegistryKeys()
{
	HKEY	paramKey;
	DWORD	paramType;
	char	cParamPath[256];
	DWORD	paramSize;
	DWORD   dwDisposition;
	LSTATUS lStatus;

	REGISTRY_KEYS srDefaultKeys;
	REGISTRY_KEYS srKeysInRegistry;

	// ERROR_FILE_NOT_FOUND == 2

	#ifdef _DEBUG_CONFIG
	DebugMsg("Verify registry keys");
	#endif

	GetKeyDefaults(&srDefaultKeys);

	memset(&srKeysInRegistry, 0, sizeof(srKeysInRegistry));

	// If the key does not exist it will be created

	GetRegistryPath(cParamPath, sizeof(cParamPath));

	strcat_s(cParamPath,sizeof(cParamPath), "\\Parameters");
	lStatus = RegCreateKeyEx(HKEY_LOCAL_MACHINE,
		cParamPath,
		0,
		NULL,
		0,
		KEY_ALL_ACCESS,
		NULL,
		&paramKey,
		&dwDisposition);

	if (lStatus != ERROR_SUCCESS)
	{
		// We had an error accessing the registry. We cannot recover from this error
		// so just exit.
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Verify config keys path error:%d",(int)lStatus);
		#endif
		return;
	}

	// we do't care if the key was created or not. The verifies below will address this.
	//if (dwDisposition == REG_CREATED_NEW_KEY)
	//{
	//} else
	//if (dwDisposition == REG_OPENED_EXISTING_KEY)
	//{
	//} else
	//{
	//}


	paramType = REG_SZ;
	paramSize = sizeof(srKeysInRegistry.tzRegistryKey);

	lStatus = RegQueryValueEx(paramKey,
						REGISTRY_TIME_ZONE_KEY,
						NULL,
						&paramType,
						(BYTE *)srKeysInRegistry.tzRegistryKey,
						&paramSize);

	if (lStatus != ERROR_SUCCESS)
	{
		// the key did not exist, set to default
		RegSetValueEx( paramKey,
							REGISTRY_TIME_ZONE_KEY,
							0,
							REG_SZ,
							(BYTE *)srDefaultKeys.tzRegistryKey,
							(DWORD)strlen(srDefaultKeys.tzRegistryKey)+1);

		strcpy_s(m_srKeys.tzRegistryKey, srDefaultKeys.tzRegistryKey);

		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Verify config keys time zone key error:%d",(int)lStatus);
		#endif
	}

	paramType = REG_SZ;
	paramSize = sizeof(srKeysInRegistry.tzDesc);

	lStatus = RegQueryValueEx(paramKey,
						REGISTRY_TIME_ZONE_DESC,
						NULL,
						&paramType,
						(BYTE *)srKeysInRegistry.tzDesc,
						&paramSize);

	if (lStatus != ERROR_SUCCESS)
	{
		// the key did not exist, set to default
		RegSetValueEx( paramKey,
							REGISTRY_TIME_ZONE_DESC,
							0,
							REG_SZ,
							(BYTE *)srDefaultKeys.tzDesc,
							(DWORD)strlen(srDefaultKeys.tzDesc)+1);

		strcpy_s(m_srKeys.tzDesc, srDefaultKeys.tzDesc);

		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Verify config keys time zone desc error:%d",(int)lStatus);
		#endif
	}

	paramType = REG_BINARY;
	paramSize = sizeof(srDefaultKeys.tzInfo);

	lStatus = RegQueryValueEx(paramKey,
						REGISTRY_TIME_ZONE_INFO,
						NULL,
						&paramType,
						(BYTE *)&srKeysInRegistry.tzInfo,
						&paramSize);

	if (lStatus != ERROR_SUCCESS)
	{
		RegSetValueEx(	paramKey,
							REGISTRY_TIME_ZONE_INFO,
							0,
							REG_BINARY,
							(BYTE *)&srDefaultKeys.tzInfo,
							sizeof(srDefaultKeys.tzInfo));

		m_srKeys.tzInfo = srDefaultKeys.tzInfo;
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Verify config keys time zone info error:%d",(int)lStatus);
		#endif
	}

	paramType = REG_DWORD;
	paramSize = sizeof(srKeysInRegistry.dateAvailable);

	lStatus = RegQueryValueEx(paramKey,
						REGISTRY_DATE_AVAIL,
						NULL,
						&paramType,
						(BYTE *)&srKeysInRegistry.dateAvailable,
						&paramSize);
	if (lStatus != ERROR_SUCCESS)
	{
		RegSetValueEx( paramKey,
							REGISTRY_DATE_AVAIL,
							0,
							REG_DWORD,
							(BYTE *)&srDefaultKeys.dateAvailable,
							sizeof(srKeysInRegistry.dateAvailable));

		m_srKeys.dateAvailable = srDefaultKeys.dateAvailable;

		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Verify config keys date avail error:%d",(int)lStatus);
		#endif
	}

	paramType = REG_DWORD;
	paramSize = sizeof(srDefaultKeys.dstEncoded);
	
	lStatus = RegQueryValueEx(paramKey,
					REGISTRY_DST_ENCODED,
					NULL,
					&paramType,
					(BYTE *)&srKeysInRegistry.dstEncoded,
					&paramSize);

	if (lStatus != ERROR_SUCCESS)
	{
		RegSetValueEx( paramKey,
							REGISTRY_DST_ENCODED,
							0,
							REG_DWORD,
							(BYTE *)&srDefaultKeys.dstEncoded,
							sizeof(srDefaultKeys.dstEncoded));

		m_srKeys.dstEncoded = srDefaultKeys.dstEncoded;
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Verify config keys dst encoded error:%d",(int)lStatus);
		#endif
	}

	paramType = REG_DWORD;
	paramSize = sizeof(srDefaultKeys.timeCodeSource);
	
	lStatus = RegQueryValueEx(paramKey,
					REGISTRY_TIME_CODE_SOURCE,
					NULL,
					&paramType,
					(BYTE *)&srKeysInRegistry.timeCodeSource,
					&paramSize);

	if (lStatus != ERROR_SUCCESS)
	{
		RegSetValueEx( paramKey,
							REGISTRY_TIME_CODE_SOURCE,
							0,
							REG_DWORD,
							(BYTE *)&srDefaultKeys.timeCodeSource,
							sizeof(srDefaultKeys.timeCodeSource));

		m_srKeys.timeCodeSource = srDefaultKeys.timeCodeSource;
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Verify config keys time code source error:%d",(int)lStatus);
		#endif
	}

	paramType = REG_DWORD;
	paramSize = sizeof(srKeysInRegistry.freewheelSynchronize);

	lStatus = RegQueryValueEx(paramKey,
					REGISTRY_FREEWHEEL_SYNC,
					NULL,
					&paramType,
					(BYTE *)&srKeysInRegistry.freewheelSynchronize,
					&paramSize);

	if (lStatus != ERROR_SUCCESS)
	{
		RegSetValueEx(	paramKey,
							REGISTRY_FREEWHEEL_SYNC,
							0,
							REG_DWORD,
							(BYTE *)&srDefaultKeys.freewheelSynchronize,
							sizeof(srDefaultKeys.freewheelSynchronize));

		m_srKeys.freewheelSynchronize = srDefaultKeys.freewheelSynchronize;
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Verify config keys freewheel sync error:%d",(int)lStatus);
		#endif
	}

	paramType = REG_DWORD;
	paramSize = sizeof(srKeysInRegistry.freewheelDisplay);

	lStatus = RegQueryValueEx(paramKey,
					REGISTRY_FREEWHEEL_DISP,
					NULL,
					&paramType,
					(BYTE *)&srKeysInRegistry.freewheelDisplay,
					&paramSize);

	if (lStatus != ERROR_SUCCESS)
	{
		RegSetValueEx(	paramKey,
							REGISTRY_FREEWHEEL_DISP,
							0,
							REG_DWORD,
							(BYTE *)&srDefaultKeys.freewheelDisplay,
							sizeof(srDefaultKeys.freewheelDisplay));

		m_srKeys.freewheelDisplay = srDefaultKeys.freewheelDisplay;
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Verify config keys freewheel display error:%d",(int)lStatus);
		#endif
	}

	paramType = REG_DWORD;
	paramSize = sizeof(srKeysInRegistry.tcLossNotify);

	lStatus = RegQueryValueEx(paramKey,
					REGISTRY_TIME_CODE_LOSS,
					NULL,
					&paramType,
					(BYTE *)&srKeysInRegistry.tcLossNotify,
					&paramSize);
	if (lStatus != ERROR_SUCCESS)
	{
		RegSetValueEx(	paramKey,
							REGISTRY_TIME_CODE_LOSS,
							0,
							REG_DWORD,
							(BYTE *)&srDefaultKeys.tcLossNotify,
							sizeof(srDefaultKeys.tcLossNotify));

		m_srKeys.tcLossNotify = srDefaultKeys.tcLossNotify;
		#ifdef _DEBUG_ERROR
		DebugMsg("!!!Verify config keys time code loss error:%d",(int)lStatus);
		#endif
	}

	// note that the remaining configuration items are undocumented or special-purpose
	// configuration and configuration load does not fail if they are missing

	//paramType = REG_DWORD;
	//paramSize = sizeof(srKeysInRegistry.setTimeAggressive);

	//lStatus = RegQueryValueEx(paramKey,
	//				REGISTRY_AGGRESSIVE_TIME_SET,
	//				NULL,
	//				&paramType,
	//				(BYTE *)&srKeysInRegistry.setTimeAggressive,
	//				&paramSize);
	//if (lStatus == ERROR_FILE_NOT_FOUND)
	//{
	//	srDefaultKeys.setTimeAggressive = 0;
	//}

	//paramType = REG_DWORD;
	//paramSize = srKeysInRegistry.useCrudeTimeSet;

	//lStatus = RegQueryValueEx(paramKey,
	//				REGISTRY_CRUDE_TIME_SET,
	//				NULL,
	//				&paramType,
	//				(BYTE *)&srKeysInRegistry.useCrudeTimeSet,
	//				&paramSize);
	//if (lStatus == ERROR_FILE_NOT_FOUND)
	//{
	//	srDefaultKeys.useCrudeTimeSet = 0;
	//}

	RegCloseKey(paramKey);
}

void CSoftwareConfig::operator=(CSoftwareConfig &dest)
{
	#ifdef _DEBUG_CONFIG
	DebugMsg("config operator =");
	#endif

	memcpy(m_srKeys.tzRegistryKey,dest.m_srKeys.tzRegistryKey,sizeof(m_srKeys.tzRegistryKey));
	memcpy(m_srKeys.tzDesc,dest.m_srKeys.tzDesc,sizeof(m_srKeys.tzDesc));
	memcpy(&m_srKeys.tzInfo,&dest.m_srKeys.tzInfo,sizeof(m_srKeys.tzInfo));
	
	m_srKeys.dstEncoded           = dest.m_srKeys.dstEncoded;
	m_srKeys.timeCodeSource       = dest.m_srKeys.timeCodeSource;
	m_srKeys.dateAvailable        = dest.m_srKeys.dateAvailable;
	m_srKeys.freewheelSynchronize = dest.m_srKeys.freewheelSynchronize;
	m_srKeys.freewheelDisplay     = dest.m_srKeys.freewheelDisplay;
	m_srKeys.tcLossNotify         = dest.m_srKeys.tcLossNotify;

	//m_srKeys.setTimeAggressive    = dest.m_srKeys.setTimeAggressive;
	//m_srKeys.useCrudeTimeSet      = dest.m_srKeys.useCrudeTimeSet;
}


//
//		Name:				::overloaded operator '=='
//
//		Description:	Compares two software configurations for equality
//
//		Returns:			TRUE if the software configurations compared are equal, 
//							otherwise FALSE.
//

int CSoftwareConfig::operator==(CSoftwareConfig &tst)
{
	#ifdef _DEBUG_CONFIG
	DebugMsg("config operator ==");
	#endif

	//if ((!strncmp(this->m_srKeys.tzRegistryKey,tst.m_srKeys.tzRegistryKey,sizeof(this->m_srKeys.tzRegistryKey))) &&
	//	(!strncmp(this->m_srKeys.tzDesc,       tst.m_srKeys.tzDesc,       sizeof(this->m_srKeys.tzDesc)))        &&
	//	CompareTimeZoneDesc(&this->m_srKeys.tzInfo,&tst.m_srKeys.tzInfo)                                &&
	//	(this->m_srKeys.dstEncoded           == tst.m_srKeys.dstEncoded)                                &&
	//	(this->m_srKeys.timeCodeSource       == tst.m_srKeys.timeCodeSource)                            &&
	//	(this->m_srKeys.dateAvailable        == tst.m_srKeys.dateAvailable)                             &&
	//	(this->m_srKeys.freewheelSynchronize == tst.m_srKeys.freewheelSynchronize)                      &&
	//	(this->m_srKeys.freewheelDisplay     == tst.m_srKeys.freewheelDisplay)                          &&
	//	(this->m_srKeys.tcLossNotify         == tst.m_srKeys.tcLossNotify)                              &&
	//	(this->m_srKeys.setTimeAggressive    == tst.m_srKeys.setTimeAggressive)                         &&
	//	(this->m_srKeys.useCrudeTimeSet      == tst.m_srKeys.useCrudeTimeSet))

	if ((!strncmp(this->m_srKeys.tzRegistryKey,tst.m_srKeys.tzRegistryKey,sizeof(this->m_srKeys.tzRegistryKey))) &&
		(!strncmp(this->m_srKeys.tzDesc,       tst.m_srKeys.tzDesc,       sizeof(this->m_srKeys.tzDesc)))        &&
		CompareTimeZoneDesc(&this->m_srKeys.tzInfo,&tst.m_srKeys.tzInfo)                                &&
		(this->m_srKeys.dstEncoded           == tst.m_srKeys.dstEncoded)                                &&
		(this->m_srKeys.timeCodeSource       == tst.m_srKeys.timeCodeSource)                            &&
		(this->m_srKeys.dateAvailable        == tst.m_srKeys.dateAvailable)                             &&
		(this->m_srKeys.freewheelSynchronize == tst.m_srKeys.freewheelSynchronize)                      &&
		(this->m_srKeys.freewheelDisplay     == tst.m_srKeys.freewheelDisplay)                          &&
		(this->m_srKeys.tcLossNotify         == tst.m_srKeys.tcLossNotify))
	{
		return(TRUE);
	}
	else
		return(FALSE);
}


//
//		Name:				::overloaded operator '!='
//
//		Description:	Compares two software configurations for inequality
//
//		Returns:			TRUE if the software configurations compared are NOT equal, 
//							otherwise FALSE.
//

int CSoftwareConfig::operator!=(CSoftwareConfig &tst)
{
	#ifdef _DEBUG_CONFIG
	DebugMsg("config operator !=");
	#endif

	//if ((!strncmp(this->m_srKeys.tzRegistryKey,tst.m_srKeys.tzRegistryKey,sizeof(this->m_srKeys.tzRegistryKey))) &&
	//	(!strncmp(this->m_srKeys.tzDesc,       tst.m_srKeys.tzDesc,       sizeof(this->m_srKeys.tzDesc)))        &&
	//	CompareTimeZoneDesc(&this->m_srKeys.tzInfo,&tst.m_srKeys.tzInfo)                                &&
	//	(this->m_srKeys.dstEncoded           == tst.m_srKeys.dstEncoded)                                &&
	//	(this->m_srKeys.timeCodeSource       == tst.m_srKeys.timeCodeSource)                            &&
	//	(this->m_srKeys.dateAvailable        == tst.m_srKeys.dateAvailable)                             &&
	//	(this->m_srKeys.freewheelSynchronize == tst.m_srKeys.freewheelSynchronize)                      &&
	//	(this->m_srKeys.freewheelDisplay     == tst.m_srKeys.freewheelDisplay)                          &&
	//	(this->m_srKeys.tcLossNotify         == tst.m_srKeys.tcLossNotify)                              &&
	//	(this->m_srKeys.setTimeAggressive    == tst.m_srKeys.setTimeAggressive)                         &&
	//	(this->m_srKeys.useCrudeTimeSet      == tst.m_srKeys.useCrudeTimeSet))

	if ((!strncmp(this->m_srKeys.tzRegistryKey,tst.m_srKeys.tzRegistryKey,sizeof(this->m_srKeys.tzRegistryKey))) &&
		(!strncmp(this->m_srKeys.tzDesc,       tst.m_srKeys.tzDesc,       sizeof(this->m_srKeys.tzDesc)))        &&
		CompareTimeZoneDesc(&this->m_srKeys.tzInfo,&tst.m_srKeys.tzInfo)                                &&
		(this->m_srKeys.dstEncoded           == tst.m_srKeys.dstEncoded)                                &&
		(this->m_srKeys.timeCodeSource       == tst.m_srKeys.timeCodeSource)                            &&
		(this->m_srKeys.dateAvailable        == tst.m_srKeys.dateAvailable)                             &&
		(this->m_srKeys.freewheelSynchronize == tst.m_srKeys.freewheelSynchronize)                      &&
		(this->m_srKeys.freewheelDisplay     == tst.m_srKeys.freewheelDisplay)                          &&
		(this->m_srKeys.tcLossNotify         == tst.m_srKeys.tcLossNotify))
	{
		return(FALSE);
	}
	else
		return(TRUE);
}


void CSoftwareConfig::SetTzDescToUTC(REGISTRY_KEYS *psrMyKeys)
{
	#ifdef _DEBUG_CONFIG
	DebugMsg("config set tz desc to utc");
	#endif

	memset(psrMyKeys->tzRegistryKey,0,sizeof(psrMyKeys->tzRegistryKey));
	strcpy_s(psrMyKeys->tzRegistryKey,sizeof(psrMyKeys->tzRegistryKey),"GMT Standard Time");

	memset(psrMyKeys->tzDesc,0,sizeof(psrMyKeys->tzDesc));
	strcpy_s(psrMyKeys->tzDesc,sizeof(psrMyKeys->tzDesc),"UTC (GMT)");
		
	psrMyKeys->tzInfo.Bias = 0L;
	memset(psrMyKeys->tzInfo.StandardName,0,sizeof(psrMyKeys->tzInfo.StandardName));
	psrMyKeys->tzInfo.StandardBias = 0L;
	memset(&psrMyKeys->tzInfo.StandardDate,0,sizeof(psrMyKeys->tzInfo.StandardDate));
	memset(psrMyKeys->tzInfo.DaylightName,0,sizeof(psrMyKeys->tzInfo.DaylightName));
	psrMyKeys->tzInfo.DaylightBias = 0L;
	memset(&psrMyKeys->tzInfo.DaylightDate,0,sizeof(psrMyKeys->tzInfo.DaylightDate));
}


//
//		Name:				::IsTzDescUTC()
//
//		Description:	Determines if the tzInfo member is functionally a UTC/GMT 
//							time zone description.
//

BOOL CSoftwareConfig::IsTzDescUTC()
{
	if (	(m_srKeys.tzInfo.Bias                == 0) &&
			(m_srKeys.tzInfo.StandardBias        == 0) &&
			(m_srKeys.tzInfo.DaylightBias        == 0) &&
			(m_srKeys.tzInfo.StandardDate.wMonth == 0) &&
			(m_srKeys.tzInfo.DaylightDate.wMonth == 0))
	{
		return(TRUE);
	}
	else
		return(FALSE);
}


//
//		Name:				::CompareTimeZoneDesc()
//
//		Description:	Compares two WCA_TIME_ZONE_INFORMATION structures for equality
//
//		Returns:			TRUE if structures are functionally equal, otherwise FALSE
//
//		Note:				Routine ignores the values of StandardName and DaylightName.
//

int CSoftwareConfig::CompareTimeZoneDesc(TIME_ZONE_INFORMATION *t1, TIME_ZONE_INFORMATION *t2)
{
	if (	(t1->Bias                       == t2->Bias) &&
			(t1->StandardBias               == t2->StandardBias) &&
			(t1->DaylightBias               == t2->DaylightBias) &&
			(t1->StandardDate.wMonth        == t2->StandardDate.wMonth) &&
			(t1->StandardDate.wDay          == t2->StandardDate.wDay) &&
			(t1->StandardDate.wYear         == t2->StandardDate.wYear) &&
			(t1->StandardDate.wHour         == t2->StandardDate.wHour) &&
			(t1->StandardDate.wMinute       == t2->StandardDate.wMinute) &&
			(t1->StandardDate.wSecond       == t2->StandardDate.wSecond) &&
			(t1->StandardDate.wMilliseconds == t2->StandardDate.wMilliseconds) &&
			(t1->DaylightDate.wMonth        == t2->DaylightDate.wMonth) &&
			(t1->DaylightDate.wDay          == t2->DaylightDate.wDay) &&
			(t1->DaylightDate.wYear         == t2->DaylightDate.wYear) &&
			(t1->DaylightDate.wHour         == t2->DaylightDate.wHour) &&
			(t1->DaylightDate.wMinute       == t2->DaylightDate.wMinute) &&
			(t1->DaylightDate.wSecond       == t2->DaylightDate.wSecond) &&
			(t1->DaylightDate.wMilliseconds == t2->DaylightDate.wMilliseconds))
	{
		return(TRUE);
	}
	else
		return(FALSE);
}


// This function is called by the API to convert the parameters received by the API
// into the parameters needed to save in the registry.
// The parameters are then saved to the registry.
//
// input: pointer to configuration values which need to be saved
//
// returns:: 0 if successful, else the parameter number which was invalid

UINT32 CSoftwareConfig::SaveConfigFromAPI(TCR1000_CONFIGURATION *pApiSetConfig)
{
	UINT32 ulReturnCode;
	int    ii;

	#ifdef _DEBUG_CONFIG
	DebugMsg("save config from API");
	#endif

	ulReturnCode = CheckConfigFromAPI(pApiSetConfig);

	if (ulReturnCode)
	{
		// one or more of the parameters failed the check, return the error(s) without saving any values
		return ulReturnCode;
	}

	GetKeyDefaults(&m_srKeys);

	m_srKeys.dstEncoded           = pApiSetConfig->DstEncoded;
	m_srKeys.freewheelDisplay     = pApiSetConfig->FreewheelDisplay;
	m_srKeys.freewheelSynchronize = pApiSetConfig->FreewheelSynchronize;
	m_srKeys.tcLossNotify         = pApiSetConfig->TimeCodeLossNotify;
	m_srKeys.timeCodeSource       = pApiSetConfig->TimeCodeSource;

	if ((m_srKeys.timeCodeSource == TC_SOURCE_DATE_INCLUDED) ||
		(m_srKeys.timeCodeSource == TC_SOURCE_DATE_INCLUDED_WITH_MC_CONTROL_BITS))
	{
		m_srKeys.dateAvailable = TRUE;
	} else
	{
		m_srKeys.dateAvailable = FALSE;
	}

	m_srKeys.tzInfo.Bias                    = pApiSetConfig->TimeZoneBiasSeconds;
	m_srKeys.tzInfo.StandardBias            = pApiSetConfig->StandardBiasSeconds;
	m_srKeys.tzInfo.StandardDate.wYear      = pApiSetConfig->StandardYear;
	m_srKeys.tzInfo.StandardDate.wMonth     = pApiSetConfig->StandardMonth;
	m_srKeys.tzInfo.StandardDate.wDayOfWeek = pApiSetConfig->StandardDayOfWeek;
	m_srKeys.tzInfo.StandardDate.wDay       = pApiSetConfig->StandardDate;
	m_srKeys.tzInfo.StandardDate.wHour      = pApiSetConfig->StandardHour;
	m_srKeys.tzInfo.StandardDate.wMinute    = pApiSetConfig->StandardMinute;
	m_srKeys.tzInfo.DaylightBias            = pApiSetConfig->DaylightBiasSeconds;
	m_srKeys.tzInfo.DaylightDate.wYear      = pApiSetConfig->DaylightYear;
	m_srKeys.tzInfo.DaylightDate.wMonth     = pApiSetConfig->DaylightMonth;
	m_srKeys.tzInfo.DaylightDate.wDayOfWeek = pApiSetConfig->DaylightDayOfWeek;
	m_srKeys.tzInfo.DaylightDate.wDay       = pApiSetConfig->DaylightDate;
	m_srKeys.tzInfo.DaylightDate.wHour      = pApiSetConfig->DaylightHour;
	m_srKeys.tzInfo.DaylightDate.wMinute    = pApiSetConfig->DaylightMinute;

	// Fill in the descriptions based on the values we received
	for (ii=0; ii<m_NumTimeZones; ii++)
	{
		if (m_srKeys.tzInfo.Bias == m_TimeZones[ii].tzInfo.Bias)
		{
			if ((m_srKeys.tzInfo.StandardBias == m_TimeZones[ii].tzInfo.StandardBias) &&
				(m_srKeys.tzInfo.DaylightBias == m_TimeZones[ii].tzInfo.DaylightBias))
			{
				if ((m_srKeys.tzInfo.StandardDate.wYear      == m_TimeZones[ii].tzInfo.StandardDate.wYear) &&
					(m_srKeys.tzInfo.StandardDate.wMonth     == m_TimeZones[ii].tzInfo.StandardDate.wMonth) &&
					(m_srKeys.tzInfo.StandardDate.wDay       == m_TimeZones[ii].tzInfo.StandardDate.wDay) &&
					(m_srKeys.tzInfo.StandardDate.wDayOfWeek == m_TimeZones[ii].tzInfo.StandardDate.wDayOfWeek) &&
					(m_srKeys.tzInfo.StandardDate.wHour      == m_TimeZones[ii].tzInfo.StandardDate.wHour) &&
					(m_srKeys.tzInfo.StandardDate.wMinute    == m_TimeZones[ii].tzInfo.StandardDate.wMinute))
				{
					if ((m_srKeys.tzInfo.DaylightDate.wYear      == m_TimeZones[ii].tzInfo.DaylightDate.wYear) &&
						(m_srKeys.tzInfo.DaylightDate.wMonth     == m_TimeZones[ii].tzInfo.DaylightDate.wMonth) &&
						(m_srKeys.tzInfo.DaylightDate.wDay       == m_TimeZones[ii].tzInfo.DaylightDate.wDay) &&
						(m_srKeys.tzInfo.DaylightDate.wDayOfWeek == m_TimeZones[ii].tzInfo.DaylightDate.wDayOfWeek) &&
						(m_srKeys.tzInfo.DaylightDate.wHour      == m_TimeZones[ii].tzInfo.DaylightDate.wHour) &&
						(m_srKeys.tzInfo.DaylightDate.wMinute    == m_TimeZones[ii].tzInfo.DaylightDate.wMinute))
					{
						// We have found the first time zone which matches what we received.
						// Save the description
						strcpy_s(m_srKeys.tzRegistryKey,sizeof(m_srKeys.tzRegistryKey),m_TimeZones[ii].tzAbbrev);
						strcpy_s(m_srKeys.tzDesc,sizeof(m_srKeys.tzDesc),m_TimeZones[ii].tzDesc);
						break;
					}
				}
			}
		}
	}

	return 0;
}


// check the configuration 
UINT32 CSoftwareConfig::CheckConfigFromAPI(TCR1000_CONFIGURATION *pApiSetConfig)
{
	UINT32 ulReturnCode;

	ulReturnCode = 0ul;

	if ((pApiSetConfig->DstEncoded < 0L) || (pApiSetConfig->DstEncoded > 1L))
		ulReturnCode |= INVALID_DST_ENCODED;

	if ((pApiSetConfig->FreewheelDisplay < 0L) || (pApiSetConfig->FreewheelDisplay > 1L))
		ulReturnCode |= INVALID_FREEWHEEL_DISPLAY;

	if ((pApiSetConfig->FreewheelSynchronize < 0L) || (pApiSetConfig->FreewheelSynchronize > 1L))
		ulReturnCode |= INVALID_FREEWHEEL_SYNCHRONIZE;

	if ((pApiSetConfig->TimeCodeLossNotify < 0L) || (pApiSetConfig->TimeCodeLossNotify > 1L))
		ulReturnCode |= INVALID_TIME_CODE_LOSS_NOTIFY;

	if ((pApiSetConfig->TimeCodeSource < 7L) || (pApiSetConfig->TimeCodeSource > 11L))
		ulReturnCode |= INVALID_TIME_CODE_SOURCE;

	if ((pApiSetConfig->TimeZoneBiasSeconds < -90000L) || (pApiSetConfig->TimeZoneBiasSeconds > 90000L))   // 25 hours max
		ulReturnCode |= INVALID_TIME_ZONE_BIAS;

	if ((pApiSetConfig->StandardBiasSeconds < -7200L) || (pApiSetConfig->StandardBiasSeconds > 7200L))   // 2 hours max
		ulReturnCode |= INVALID_STANDARD_BIAS;

	if (pApiSetConfig->StandardYear)
	{
		if ((pApiSetConfig->StandardYear < 1900L) || (pApiSetConfig->StandardYear > 2200L))
			ulReturnCode |= INVALID_STANDARD_YEAR;
	}

	if ((pApiSetConfig->StandardMonth < 0L) || (pApiSetConfig->StandardMonth > 12L))
		ulReturnCode |= INVALID_STANDARD_MONTH;

	if ((pApiSetConfig->StandardDayOfWeek < 0L) || (pApiSetConfig->StandardDayOfWeek > 7L))
		ulReturnCode |= INVALID_STANDARD_DAY_OF_WEEK;

	if ((pApiSetConfig->StandardDate < 0L) || (pApiSetConfig->StandardDate > 31L))
		ulReturnCode |= INVALID_STANDARD_DATE;

	if ((pApiSetConfig->StandardHour < 0L) || (pApiSetConfig->StandardHour > 23L))
		ulReturnCode |= INVALID_STANDARD_HOUR;

	if ((pApiSetConfig->StandardMinute < 0L) || (pApiSetConfig->StandardMinute > 59L))
		ulReturnCode |= INVALID_STANDARD_MINUTE;

	if ((pApiSetConfig->DaylightBiasSeconds < -7200L) || (pApiSetConfig->DaylightBiasSeconds > 7200L))   // 2 hours max
		ulReturnCode |= INVALID_DAYLIGHT_BIAS;

	if (pApiSetConfig->DaylightYear)
	{
		if ((pApiSetConfig->DaylightYear < 1900L) || (pApiSetConfig->DaylightYear > 2200L))
			ulReturnCode |= INVALID_DAYLIGHT_YEAR;
	}

	if ((pApiSetConfig->DaylightMonth < 0L) || (pApiSetConfig->DaylightMonth > 12L))
		ulReturnCode |= INVALID_DAYLIGHT_MONTH;

	if ((pApiSetConfig->DaylightDayOfWeek < 0L) || (pApiSetConfig->DaylightDayOfWeek > 7L))
		ulReturnCode |= INVALID_DAYLIGHT_DAY_OF_WEEK;

	if ((pApiSetConfig->DaylightDate < 0L) || (pApiSetConfig->DaylightDate > 31L))
		ulReturnCode |= INVALID_DAYLIGHT_DATE;

	if ((pApiSetConfig->DaylightHour < 0L) || (pApiSetConfig->DaylightHour > 23L))
		ulReturnCode |= INVALID_DAYLIGHT_HOUR;

	if ((pApiSetConfig->DaylightMinute < 0L) || (pApiSetConfig->DaylightMinute > 59L))
		ulReturnCode |= INVALID_DAYLIGHT_MINUTE;

	return ulReturnCode;
}


// This function is called by the API to convert the registry parameters into those
// used by the API.
// The converted parameters are then returned to the API.
//
// input: pointer to where to save the configuration values
//
// returns:: none

void CSoftwareConfig::GetConfigForAPI(TCR1000_CONFIGURATION *pApiGetConfig)
{
	#ifdef _DEBUG_CONFIG
	DebugMsg("get config for API");
	#endif

	pApiGetConfig->DstEncoded           = m_srKeys.dstEncoded;
	pApiGetConfig->FreewheelDisplay     = m_srKeys.freewheelDisplay;
	pApiGetConfig->FreewheelSynchronize = m_srKeys.freewheelSynchronize;
	pApiGetConfig->TimeCodeLossNotify   = m_srKeys.tcLossNotify;
	pApiGetConfig->TimeCodeSource       = m_srKeys.timeCodeSource;

	pApiGetConfig->TimeZoneBiasSeconds  = m_srKeys.tzInfo.Bias;
	pApiGetConfig->StandardBiasSeconds  = m_srKeys.tzInfo.StandardBias;
	pApiGetConfig->StandardYear         = m_srKeys.tzInfo.StandardDate.wYear;
	pApiGetConfig->StandardMonth        = m_srKeys.tzInfo.StandardDate.wMonth;
	pApiGetConfig->StandardDayOfWeek    = m_srKeys.tzInfo.StandardDate.wDayOfWeek;
	pApiGetConfig->StandardDate         = m_srKeys.tzInfo.StandardDate.wDay;
	pApiGetConfig->StandardHour         = m_srKeys.tzInfo.StandardDate.wHour;
	pApiGetConfig->StandardMinute       = m_srKeys.tzInfo.StandardDate.wMinute;
	pApiGetConfig->DaylightBiasSeconds  = m_srKeys.tzInfo.DaylightBias;
	pApiGetConfig->DaylightYear         = m_srKeys.tzInfo.DaylightDate.wYear;
	pApiGetConfig->DaylightMonth        = m_srKeys.tzInfo.DaylightDate.wMonth;
	pApiGetConfig->DaylightDayOfWeek    = m_srKeys.tzInfo.DaylightDate.wDayOfWeek;
	pApiGetConfig->DaylightDate         = m_srKeys.tzInfo.DaylightDate.wDay;
	pApiGetConfig->DaylightHour         = m_srKeys.tzInfo.DaylightDate.wHour;
	pApiGetConfig->DaylightMinute       = m_srKeys.tzInfo.DaylightDate.wMinute;

}

//
//	Name:			::GetRegistryTimeZones()
//
//	Description:	Read all time zone information from registry into
//					array of TIME_ZONE_DESC
//

BOOL CSoftwareConfig::GetRegistryTimeZones(void)
{
	HKEY		enumKey, tzKey;
	char		tzKeyName[512], buffer[512];
	DWORD		tzKeyNameSize, tzDescKeyType, bufferSize;
	FILETIME	lastWrite;
	char		tzPath[256];

	if (GetVersion() & 0x80000000)
		strcpy_s(tzPath,sizeof(tzPath),"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Time Zones");
	else
		strcpy_s(tzPath,sizeof(tzPath),"SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones");

	RegOpenKeyEx(HKEY_LOCAL_MACHINE,tzPath,0,KEY_READ,&enumKey);

	m_NumTimeZones = 0;
	tzKeyNameSize = sizeof(tzKeyName);

	while(RegEnumKeyEx(enumKey,m_NumTimeZones,tzKeyName,&tzKeyNameSize,NULL,NULL,NULL,&lastWrite) == ERROR_SUCCESS)
	{
		memset(m_TimeZones[m_NumTimeZones].tzAbbrev, 0, sizeof(m_TimeZones[m_NumTimeZones].tzAbbrev));

		strncpy_s(m_TimeZones[m_NumTimeZones].tzAbbrev,sizeof(m_TimeZones[m_NumTimeZones].tzAbbrev),tzKeyName,255);

		RegOpenKeyEx(enumKey,tzKeyName,0,KEY_READ,&tzKey);

		tzDescKeyType = REG_SZ;
		bufferSize = sizeof(buffer);

		RegQueryValueEx(tzKey,"Display",NULL,&tzDescKeyType,(BYTE *)buffer,&bufferSize);

		memset(m_TimeZones[m_NumTimeZones].tzDesc, 0, sizeof(m_TimeZones[m_NumTimeZones].tzDesc));

		strncpy_s(m_TimeZones[m_NumTimeZones].tzDesc,sizeof(m_TimeZones[m_NumTimeZones].tzDesc),buffer,255);

		tzDescKeyType = REG_BINARY;
		bufferSize = sizeof(buffer);

		RegQueryValueEx(tzKey,"TZI",NULL,&tzDescKeyType,(BYTE *)buffer,&bufferSize);

		memcpy(&m_TimeZones[m_NumTimeZones].tzInfo,buffer,sizeof(TIME_ZONE_DESC));

		RegCloseKey(tzKey);
		tzKeyNameSize = sizeof(tzKeyName);

		m_NumTimeZones++;

		RegEnumKeyEx(enumKey,m_NumTimeZones,tzKeyName,&tzKeyNameSize,NULL,NULL,NULL,&lastWrite);
	}

	// sort time zones according to bias

	qsort(&m_TimeZones[0],m_NumTimeZones,sizeof(TIME_ZONE_DESC),tzCompare);

	return TRUE;
}



//
//	Name:			tzCompare()
//
//	Description:	qsort comparison routine for sorting arrays of
//					TIME_ZONE_DESC structures by bias offset
//

int tzCompare(const void *arg1, const void *arg2)
{
	if (((TIME_ZONE_DESC *)arg1)->tzInfo.Bias < ((TIME_ZONE_DESC *)arg2)->tzInfo.Bias)
		return -1;

	if (((TIME_ZONE_DESC *)arg1)->tzInfo.Bias > ((TIME_ZONE_DESC *)arg2)->tzInfo.Bias)
		return 1;

	return 0;
}