#pragma once


#include "tzlib.h"
//#include <wtypes.h>

typedef struct _REGISTRY_TIMEZONE_INFO
{
	LONG				Bias;
	LONG				StandardBias;
	LONG				DaylightBias;
	SYSTEMTIME			StandardDate;
	SYSTEMTIME			DaylightDate;

} REGISTRY_TIMEZONE_INFO, *PREGISTRY_TIMEZONE_INFO;


#define REGISTRY_SOURCE_NAME			"TCR1000SyncConfiguration"

// time code products

//#define TC_PRODUCT_TCR200				0x00000001L
//#define TC_PRODUCT_TCR200I			0x00000002L
//#define TC_PRODUCT_TCR200E			0x00000003L
//#define TC_PRODUCT_TCR500				0x00000004L
//#define TC_PRODUCT_TCR500E			0x00000005L
//#define TC_PRODUCT_TCR500PCI			0x00000006L
//#define TC_PRODUCT_TCR1000			0x00000007L
#define TC_PRODUCT_TCR1000_REV_A		0x00000008L
#define TC_PRODUCT_TCR1000_X64_REV_A	0x00000009L
#define TC_PRODUCT_UNDEFINED			0xffffffffL

// time code sources

//#define TC_SOURCE_GPS100				0x00000001L	// GPS-100 master clock
//#define TC_SOURCE_OTHER				0x00000002L	// ...any other
//#define TC_SOURCE_GPS200				0x00000003L	// GPS-200/200A
//#define TC_SOURCE_TCI232				0x00000004L // TCI-232
//#define TC_SOURCE_TCI232A				0x00000005L	// TCI-500/TCI-232A
//#define TC_SOURCE_TCO100				0x00000006L // TCO-100
// The defines above are no longer used. Instead define the
// information included with the time code.
#define TC_SOURCE_DATE_NOT_INCLUDED                   0x00000007L
#define TC_SOURCE_DATE_INCLUDED                       0x00000008L
#define TC_SOURCE_DATE_INCLUDED_WITH_MC_CONTROL_BITS  0x00000009L
#define TC_SOURCE_DATE_IRIG_1344                      0x0000000AL // not implemented
#define TC_SOURCE_DATE_SMPTE_309M                     0x0000000BL // not implemented
#define TC_SOURCE_UNDEFINED				              0xffffffffL // undefined

// time code types

// now defined in IntAptDefs.h
//#define TC_TYPE_SMPTE					0x00000001L
//#define TC_TYPE_IRIG					0x00000002L
//#define TC_TYPE_UNDEFINED				0x00008000L

// time code sub-types

//#define TC_SUBTYPE_SMPTE30FPS			0x00010000L
//#define TC_SUBTYPE_EBU25FPS			0x00020000L
//#define TC_SUBTYPE_FILM24FPS			0x00040000L
//#define TC_SUBTYPE_IRIG_B1			0x00080000L
//#define TC_SUBTYPE_IRIG_B				0x00100000L
//#define TC_SUBTYPE_UNDEFINED			0x80000000L

//********************************************************************************
//
// Data structures and definitions used by TCRSync's named pipe and DDE
// interprocess communications.
//
// The Windows NT/2000 TCRSync services utilize the named pipe mechanism.
// The Windows 9x/Me TCRSync services utilize the DDE mechanism.
//
//********************************************************************************

#define TCRSYNC_IPC_NAME				"TCRSYNC_PIPE"
#define TCRSYNC_IPC_SYNCHISTMAX			10

// get synchronization service version
#define TCRSYNC_IPCCMD_VERSION			20
#define TCRSYNC_IPCCMD_VERSION_STR		"TCRSYNC_VERSION"
// get synchronization history
#define TCRSYNC_IPCCMD_SYNCHIST			21
#define TCRSYNC_IPCCMD_SYNCHIST_STR		"TCRSYNC_SYNCHIST"
// get current UTC time from service
#define TCRSYNC_IPCCMD_CURTIME			22
#define TCRSYNC_IPCCMD_CURTIME_STR		"TCRSYNC_CURTIME"
// get TCRSync operation info
#define TCRSYNC_IPCCMD_INFO				23
#define TCRSYNC_IPCCMD_INFO_STR			"TCRSYNC_INFO"
// causes synchronization service to set 60Hz interrupt frequency
#define TCRSYNC_IPCCMD_SETINTFREQ		24
#define TCRSYNC_IPCCMD_SETINTFREQ_STR	"TCRSYNC_SETINTFREQ"

#define TCRSYNC_IPCCMD_RECALIBRATE		25
#define TCRSYNC_IPCCMD_RECALIBRATE_STR	("TCRSYNC_RECALIBRATE")

// parameters for TCRSYNC_IPCCMD_SETINTFREQ

#define INTERRUPT_FREQ_DEFAULT			100	// 1/60Hz interrupt from time code reader
#define INTERRUPT_FREQ_PERSEC			101	// 1Hz interrupt from time code reader

// time synchronization status flags

#define CURTIME_HEALTHY				0
#define CURTIME_UNHEALTHY			1
#define CURTIME_INVALID				2

// time synchronization failure detail status flags

#define REASON_NOREASON			0	// indicates no reason/no condition
#define REASON_TIMELOSS			1	// time is unhealthy or invalid due to no time available
#define REASON_SERIOUSTIMELOSS	2	// serious time code loss has occured
#define REASON_GPSLOSS			3	// time code reference is freewheeling
									// GPS-200A/MCR1000/MCR5000 operation statuses when Masterclock, Inc proprietary control
									// bits are included in SMPTE time code, Leitch date encoding)
#define REASON_SERIOUSGPSLOSS	4	// serious satellite acquisition problem
									// GPS-200A/MCR1000/MCR5000 operation statuses when Masterclock, Inc proprietary control
									// bits are included in SMPTE time code, Leitch date encoding)

#pragma pack(push, 4)  // force word structure packing so that all programs using
					   // this header are guaranteed the same size objects

// Pipe transaction control structure:
// -----------------------------------

typedef struct
{
	DWORD	cmd;						// command
	DWORD	length;						// length of data to be expected (0 if none)
} PIPE_CONTROL, *PPIPE_CONTROL;


// Software version information:
// -----------------------------

typedef struct
{
	DWORD	verMaj;						// TCRSync major version number
	DWORD	verMin;						// TCRSync minor version number
} TCRSYNC_IPC_VERSION, *PTCRSYNC_IPC_VERSION;


// Last-sync time information:
// ---------------------------
// The time provided here is a record of the last TCRSYNC_PIPE_SYNCHISTMAX time
// synchronizations that occured on the NT system running TCRSync.  The numSyncRecords
// member indicates how many of the record entries are valid (filled in) starting with
// entry 0.  Entry 0 represents the most recent time synchronization, etc.
// (This information is not for time synchronization use!)

typedef struct
{
	DWORD		numSyncRecords;			// number entries in synchronization record
	WCATIME		decodedTime[TCRSYNC_IPC_SYNCHISTMAX];
	WCATIME		utcTime[TCRSYNC_IPC_SYNCHISTMAX];
	SYSTEMTIME	systemTime[TCRSYNC_IPC_SYNCHISTMAX]; // before synchronization
} TCRSYNC_IPC_SYNCHIST, *PTCRSYNC_IPC_SYNCHIST;


// TCRSync current time information:
// ---------------------------------
// The current time as derived by TCRSync from the time code source.  If the time code source
// does not provide date information date it is derived from the NT system running TCRSync
// (but the date members of 'rawTime' will be set to zero).
//
// If 'projectedTime' is non-zero it indicates the number of seconds since the last interrupt
// that provided TCRSync with time.  The time contained in the 'rawTime' and 'utcTime' members
// will projected forward by that number of seconds. ('wMilliseconds' member of each of the
// two time objects is also correctly projected.)  Projected time is considered accurate for
// time synchronization over this named pipe mechanism unless timeStatus is flagged other
// than healthy.
//
//  The 'localTime' member of this object is always the current local system time on the
// system running TCRSync and is not guaranteed to be synchronized with the
// 'rawTime'/'utcTime' members and should never be used for time synchronization
// purposes.  ONLY THE 'utcTime' MEMBER SHOULD BE USED FOR TIME SYNCHRONIZATION.

typedef struct {
	BYTE			timeStatus;		// status of provided time
	BYTE			extStatus;		// extended explanation of status (if timeStatus not healthy)
	DWORD			projectedTime;	// time given in UTC and local time members is projected
	WCATIME			rawTime;		// time as read from time code reader at last interrupt
	WCATIME			utcTime;		// current UTC time (use this for time sync)
	SYSTEMTIME		localTime;		// current time on host running TCRSync

	DWORD			reserved1;
	DWORD			reserved2;
} TCRSYNC_IPC_CURTIME, *PTCRSYNC_IPC_CURTIME;


// TCRSync operation information:
// ------------------------------
// Operation information and event counts.  All event counts are relative to the last time
// TCRSync was started.



//typedef struct {
//	SYSTEMTIME	startTime;			// local clock time when TCRSync started
//	DWORD		syncCount;			// number of synchronizations that have been applied
//	DWORD		rCompare;			// rolling time comparison failures
//	DWORD		seriousTimeLoss;	// serious time losses detected
//	DWORD		seriousGpsLoss;		// serious GPS losses detected
//	DWORD		TimeCodeType;		// IRIG, SMPTE
//	uTimeCode   rawTc;				// raw time code
//	DWORD		reserved1;
//	DWORD		reserved2;
//	DWORD		reserved3;
//	DWORD		reserved4;
//} TCRSYNC_IPC_INFO, *PTCRSYNC_IPC_INFO;

#define INVALID_DST_ENCODED           0x00000001UL
#define INVALID_FREEWHEEL_DISPLAY     0x00000002UL
#define INVALID_FREEWHEEL_SYNCHRONIZE 0x00000004UL
#define INVALID_TIME_CODE_LOSS_NOTIFY 0x00000008UL
#define INVALID_TIME_CODE_SOURCE      0x00000010UL
#define INVALID_TIME_ZONE_BIAS        0x00000020UL
#define INVALID_STANDARD_BIAS         0x00000040UL
#define INVALID_STANDARD_YEAR         0x00000080UL
#define INVALID_STANDARD_MONTH        0x00000100UL
#define INVALID_STANDARD_DAY_OF_WEEK  0x00000200UL
#define INVALID_STANDARD_DATE         0x00000400UL
#define INVALID_STANDARD_HOUR         0x00000800UL
#define INVALID_STANDARD_MINUTE       0x00001000UL
#define INVALID_DAYLIGHT_BIAS         0x00002000UL
#define INVALID_DAYLIGHT_YEAR         0x00004000UL
#define INVALID_DAYLIGHT_MONTH        0x00008000UL
#define INVALID_DAYLIGHT_DAY_OF_WEEK  0x00010000UL
#define INVALID_DAYLIGHT_DATE         0x00020000UL
#define INVALID_DAYLIGHT_HOUR         0x00040000UL
#define INVALID_DAYLIGHT_MINUTE       0x00080000UL

typedef struct {
	INT32 DstEncoded;				// ==0, time code never offset by daylight time
									// ==1, time code offset by daylight time during daylight period
	INT32 FreewheelDisplay;			// ==0, time code does not contain Masterclock, Inc proprietary freewheeling indicator in SMPTE frame
									// ==1, when freewheeling, time code will indicate such using Masterclock, Inc proprietary lock indicator in SMPTE frame 
	INT32 FreewheelSynchronize;		// ==0, when freewheeling (FreewheelDisplay must ==1 to know when freewheeling), do not set system time
									// ==1, when freewheeling (FreewheelDisplay must ==1 to know when freewheeling), set system time to incoming time code
	INT32 TimeCodeLossNotify;		// ==0, do not generate alarm when time code lock is lost
									// ==1, generate alarm when time code lock is lost
	INT32 TimeCodeSource;			// ==7, date not included in time code
									// ==8, date included in time code
									// ==9, date and Masterclock, Inc proprietary control bits included in SMPTE time code
									// ==10, time code uses IRIG 1344 encoding (NOT IMPLEMENTED)
									// ==11, time code uses SMPTE 309M encoding (NOT IMPLEMENTED)
	INT32 TimeZoneBiasSeconds;		// time zone bias, in seconds
	INT32 StandardBiasSeconds;		// standard bias, in seconds (usually zero)
	INT32 StandardYear;				// zero for relative standard time, else year for absolute standard date (usually zero)
	INT32 StandardMonth;			// month standard time begins
	INT32 StandardDayOfWeek;		// day of week standard time begins, SUN==1, MON==2, TUE==3, etc.
	INT32 StandardDate;             // date of month standard time begins
	INT32 StandardHour;				// hour of day standard time begins
	INT32 StandardMinute;           // minute of hour standard time begins
	INT32 DaylightBiasSeconds;		// daylight bias, in seconds (usually zero)
	INT32 DaylightYear;				// zero for relative daylight time, else year for absolute daylight date (usually zero)
	INT32 DaylightMonth;			// month daylight time begins
	INT32 DaylightDayOfWeek;		// day of week daylight time begins, SUN==1, MON==2, TUE==3, etc.
	INT32 DaylightDate;				// date of month daylight time begins
	INT32 DaylightHour;				// hour of day daylight time begins
	INT32 DaylightMinute;			// minute of hour daylight time begins
} TCR1000_CONFIGURATION;

#pragma pack(pop)

 
